/* 
 * cfg.cc
 */

#include <stdio.h>

#include "cfg.h"

/* #define FREAD_TOKEN_LIMIT 4096 */
#define FREAD_TOKEN_LIMIT 2048
int read_token_int(FILE * st, char delimiter)
{
  char s[FREAD_TOKEN_LIMIT]; 
  int del = (int) delimiter;
  int ok = 0;
  int i;
  for (i = 0; i < FREAD_TOKEN_LIMIT - 1; i++) {
    int c = fgetc (st);
    if (c == del) {
      s[i] = del; 
      i++; 
      s[i] = 0; ok = 1; break;
    } else if (c == EOF) {
      s[i] = 0; ok = 1; break;
    } else {
      s[i] = c;
    }
  }
  if (ok == 0) {
    fprintf(stderr, 
	    "Error: read_token_int buffer overflow (delimiter = %d)\n",
	    delimiter);
    st_wg_exit((void *)1);
    return 0;
  } else {
    return atoi (s);
  }
}

st_list_t(st_list_t(int)) corpus_reader(char * filename, st_region_t rg)
{
  FILE * s = fopen(filename, "r");
  if (s == 0) {
    fprintf (stderr, "can't open %s\n", filename);
    perror("corpus_reader");
    st_wg_exit((void *)1);
  }
  st_list_t(st_list_t(int)) cps = 0;
  while (!feof(s)) {
    st_list_t(int) sen = 0; int delimit = 0;
    while (!delimit) {
      int word = read_token_int (s, CHAR_SPACE);
      if (word == -1) delimit = 1;
      else sen = cons(word, sen, rg);
    }
    read_token_int(s, CHAR_NEWLINE); // skip newline
    cps = cons(reverse(sen, rg), cps, rg);
  }
  fclose(s);
  return cps;
}

enum verbose_level_t {
    verbose_level_none,
    verbose_level_count,
    verbose_level_show,
    verbose_level_last,		// unused
};

void cfg_main(char * grammar_file, char * corpus_file, verbose_level_t vl)
{
  st_region_t g_rg = MM_MAKE_REGION();
  grammar * grm = make_grammar(grammar_file, g_rg);
  cky_matrix_t cky_mat = make_cky_matrix(g_rg);
  clear_cky_matrix(cky_mat, MAX_CKY_MAT_SIZE - 1);
  st_list_t(edge_t) * G = make_result_matrix(g_rg);
  clear_result_matrix(G, MAX_CKY_MAT_SIZE - 1);
  st_list_t(st_list_t(int)) corpus = corpus_reader(corpus_file, g_rg);

  st_list_t(st_list_t(int)) sen;
  int i; 
  int n_total_syms = 0;
  int total_time = 0;

  st_region_t p_rg = MM_MAKE_REGION();
  for (sen = corpus, i = 0; sen; sen = sen->cdr, i++) {

    st_list_t(int) s = sen->car;
    int l = length(s);
    printf ("parsing %d th sentence (%d words) ... ", i, l); 
    fflush(stdout);
    long t0 = st_current_time_ms(); long t1, t2;
	
    cky_parse(s, grm, cky_mat, G, p_rg);

    t1 = st_current_time_ms();
    
    if (vl >= verbose_level_count) {
	int st = count_syms(cky_mat, l);
	n_total_syms += st;
	printf ("%d syms\n", st);
    }
    if (vl >= verbose_level_show) {
	show_cky_matrix(cky_mat, l);
	show_count_syms(cky_mat, l);
    } 

    clear_cky_matrix(cky_mat, l);
    clear_result_matrix(G, l);
    MM_RESET_REGION(p_rg);
    MM_GC_POINT();

    t2 = st_current_time_ms();
    total_time += (t2 - t0);
    printf ("done (%d msec)\n", t2 - t0);
  }

  MM_RESET_REGION(g_rg);
  MM_GC_POINT();
#if UNDERLYING_MM_IS_SGC
  printf ("%d GCs\n", GC_gc_no);
#endif
  if (vl >= verbose_level_count) {
    printf ("%d edges\n", n_total_syms);
  } else {
    printf ("did not count edges\n");
  }
}

int st_main (int argc, char **argv)
{
  if (argc != 4) {
    printf("usage: %s grammar corpus verbose_level(=0,1,2)\n"
	   "example: %s M corpus 0\n", argv[0], argv[0]);
    st_wg_exit((void *)1);
  }
  printf("use large stack size (> 4m) for this program\n");

  char * grammar_file = argv[1];
  char * corpus_file = argv[2];
  int _vl = atoi(argv[3]);
  verbose_level_t vl;
  if (_vl < (long)verbose_level_last) {
    vl = (verbose_level_t)_vl;
  } else {
    printf("verbose_level must be <= %d\n", verbose_level_last - 1);
    st_wg_exit((void *)1);
  }
  MM_SETUP(/* GC_free_space_divisor = */ 2);
  cfg_main(grammar_file, corpus_file, vl);
  return 0;
}

