/*
 * cky.cc
 */

#include "cfg.h"

void foo ()
{
  cons (0, (st_list_t(int))0);
}

cky_matrix_t make_cky_matrix(st_region_t rg)
{
  int size = MAX_CKY_MAT_SIZE * MAX_CKY_MAT_SIZE;
  cky_matrix_t mat = new (rg) cky_matrix_elem_type[size];

  int g; int i, j; int n = MAX_CKY_MAT_SIZE - 1;
  for (g = 1; g <= n; g++) {
    for (i = 0, j = i + g; j <= n; i++, j++) {
      MREF(mat, i, j).clear(0);
    }
  }
  return mat;
}

result_matrix_t make_result_matrix(st_region_t rg)
{
  int size = MAX_CKY_MAT_SIZE * MAX_CKY_MAT_SIZE;
  result_matrix_t ge = new (rg) st_list_t(edge_t)[size];

  int g; int i, j; int n = MAX_CKY_MAT_SIZE - 1;
  for (g = 1; g <= n; g++) {
    for (i = 0, j = i + g; j <= n; i++, j++) {
      MREF(ge, i, j) = 0;
    }
  }
  return ge;
}

st_list_t(int) remove_prefix(int n, st_list_t(int) l)
{
  int i; st_list_t(int) p;
  for (i = 0, p = l; i < n; i++, p = p->cdr) {
  }
  return p;
}

void cky_parse_aux(int i, int j, st_list_t(int) s, 
		   grammar * grm, cky_matrix_t cky_mat, result_matrix_t G,
		   st_region_t rg)
{
  st_list_t(int) Pij;
  int k;

  if (i + 1 == j) {
    st_list_t(edge_t) * g = &(MREF(G, i, j));
    st_list_t(int) x = cons(s->car, (st_list_t(int))0, rg);
    Pij = grm->apply_unary_rules(x, 0, g, rg);
    // printf("[%d,%d] <-- answer\n", i, j);
    MREF(cky_mat, i, j).put(Pij);
  } else {
    for (k = i + 1; k < j; k++) {
#if MTCAMP
      if (MREF(cky_mat, i, k).acquire()) mtc_raw_fork {
	ST_POLLING();
	cky_parse_aux(i, k, s, grm, cky_mat, G, rg);
      }
      if (MREF(cky_mat, k, j).acquire()) mtc_raw_fork {
	ST_POLLING();
	cky_parse_aux(k, j, remove_prefix(k - i, s), grm, cky_mat, G, rg);
      }
#elif STHREADS
      if (MREF(cky_mat, i, k).acquire()) {
	ST_POLLING();
	ST_THREAD_CREATE(cky_parse_aux(i, k, s, grm, cky_mat, G, rg));
      }
      if (MREF(cky_mat, k, j).acquire()) {
	ST_POLLING();
	ST_THREAD_CREATE(cky_parse_aux(k, j, remove_prefix(k - i, s), 
				       grm, cky_mat, G, rg));
      }
#else
      if (MREF(cky_mat, i, k).acquire()) {
	cky_parse_aux(i, k, s, grm, cky_mat, G, rg);
      }
      if (MREF(cky_mat, k, j).acquire()) {
	cky_parse_aux(k, j, remove_prefix(k - i, s), grm, cky_mat, G, rg);
      }
#endif
    }
  
    Pij = 0;
    st_list_t(edge_t) * g = &(MREF(G, i, j));
    for (k = i + 1; k < j; k++) {
      ST_POLLING();
      // printf("[%d,%d] tries to read [%d,%d]\n", i, j, i, k);
      st_list_t(int) Pik = MREF(cky_mat, i, k).get();
      // printf("[%d,%d] read [%d,%d]\n", i, j, i, k);
      // printf("[%d,%d] tries to read [%d,%d]\n", i, j, k, j);
      st_list_t(int) Pkj = MREF(cky_mat, k, j).get();
      // printf("[%d,%d] read [%d,%d]\n", i, j, k, j);
      Pij = grm->apply_binary_rules(Pik, Pkj, k, Pij, g, rg);
    }
    Pij = grm->apply_unary_rules(Pij, Pij, g, rg);
    // printf("[%d,%d] <-- answer\n", i, j);
    MREF(cky_mat, i, j).put(Pij);
  }

  ST_POLLING();
}

void print_int(int x)
{
    printf("%d ", x);
}

void cky_parse(st_list_t(int) sentence, 
	       grammar * grm, cky_matrix_t cky_mat, result_matrix_t G, 
	       st_region_t rg)
{
  int n = length(sentence);
  
  if (n > MAX_CKY_MAT_SIZE) {
    fprintf (st_errout, "sentence too long (%d elems) > %d\n", 
	     n, MAX_CKY_MAT_SIZE);
    st_list_for_each(print_int, sentence);
    st_wg_exit((void *)1);
  }
  
  int a = MREF(cky_mat, 0, n).acquire();
  st_assert(a);
  cky_parse_aux(0, n, sentence, grm, cky_mat, G, rg);
  
  ST_POLLING();
}

void clear_cky_matrix(cky_matrix_t cky_mat, int n)
{
  int g;
  for (g = 1; g <= n; g++) {
    int i, j;
    for (i = 0, j = i + g; j <= n; i++, j++) {
      MREF(cky_mat, i, j).clear(0);
    }
  }
}

void clear_result_matrix(result_matrix_t G, int n)
{
  int g;
  for (g = 1; g <= n; g++) {
    int i, j;
    for (i = 0, j = i + g; j <= n; i++, j++) {
      MREF(G, i, j) = 0;
    }
  }
}

void show_cky_matrix_1(cky_matrix_t cky_mat, int i, int j)
{
  printf("P[%d,%d] = ", i, j);
  st_list_for_each(print_int, MREF(cky_mat, i, j).get());
  printf ("\n");
}
    
void show_cky_matrix(cky_matrix_t cky_mat, int n)
{
  int i, j, g;
  for (g = 1; g <= n; g++) {
    for (i = 0, j = i + g; j <= n; i++, j++) {
      show_cky_matrix_1(cky_mat, i, j);
    }
  }
}

int count_syms_1(cky_matrix_t cky_mat, int i, int j)
{
  return length(MREF(cky_mat, i, j).get());
}

int count_syms(cky_matrix_t cky_mat, int n)
{
  int i, j, g;
  int total = 0;
  
  for (g = 1; g <= n; g++) {
    for (i = 0, j = i + g; j <= n; i++, j++) {
      int s = count_syms_1(cky_mat, i, j);
      // printf ("#P[%d,%d] = %d\n", i, j, s);
      total += s;
    }
  }
  return total;
}

void show_count_syms(cky_matrix_t cky_mat, int n)
{
  int i, j, g;
  
  for (g = 1; g <= n; g++) {
    for (i = 0, j = i + g; j <= n; i++, j++) {
      int s = count_syms_1(cky_mat, i, j);
      printf ("#P[%d,%d] = %d\n", i, j, s);
    }
  }
}


