/*
 * nbody.cc
 */
#include <sys/types.h>
#include <sys/time.h>

#include "def.h"

#if COUNT_NEWTONS
extern long n_newton;
#endif

void simulate_a_step_(particle ** particles, int n_particles, 
		     t_real dt, 
		     st_region_t rg, st_region_t mp_rg)
{
  long t0, t1, t2, t3, t4;
  space * tree;
  
  t0 = st_current_time_ms();
  tree = generate_tree(particles, n_particles, rg);
  t1 = st_current_time_ms();
  mass_momentum mm = tree->set_mass_and_cg(rg);
  t2 = st_current_time_ms();
  printf("%d nodes in the tree\n", mm.n_nodes);
  set_accels(particles, n_particles, tree, rg);
  t3 = st_current_time_ms();
  move_particles(particles, n_particles, dt, mp_rg);
  t4 = st_current_time_ms();
  printf("generate-tree: %d msec\n", t1 - t0);
  printf("set-mass-and-cg: %d msec\n", t2 - t1);
  printf("set-accels: %d msec\n", t3 - t2);
  printf("move-particles: %d msec\n", t4 - t3);
  printf("sum: %d msec\n", t4 - t0);
#if COUNT_NEWTONS
  printf ("%d newtons\n", st_fetch_long(&n_newton));
#endif
}

void simulate_a_step(particle ** particles, int n_particles, 
		     t_real dt, 
		     st_region_t rg, st_region_t mp_rg);

void dump_particles (particle **, int);

void simulate_steps_(particle ** particles, int n_particles, 
		    int n_steps, t_real dt)
{
  st_region_t rg0 = MM_MAKE_REGION();
  st_region_t rg1 = MM_MAKE_REGION();
  t_real dt1 = dt; /* endo */
  for (int i = 0; i < n_steps; i++) {
#if 1/*endo debug*/
    printf("[A] dt %g dt1 %g\n", dt, dt1);
#endif
    simulate_a_step(particles, n_particles, dt1, rg0, rg1);
#if 1/*endo debug*/
    printf("[B] dt %g dt1 %g\n", dt, dt1);
#endif
    ST_POLLING();
#if 1/*endo debug*/
    printf("[C] dt %g dt1 %g\n", dt, dt1);
#endif
    st_region_t tmp = rg0;
    rg0 = rg1;
    rg1 = tmp;
    MM_RESET_REGION(rg1);
    MM_GC_POINT();
#if 1/*endo debug*/
    printf("[D] dt %g dt1 %g\n", dt, dt1);
#endif
    printf("%d : stack size after %dth step = %d\n", 
	   tls(thread_id), i, st_stack_used_bytes());
  }
}

void simulate_steps(particle ** particles, int n_particles, 
		    int n_steps, t_real dt);

void nbody_main (int n_particles, int n_steps, t_real dt, int dump_p)
{
  printf("nbody_main\n");
  st_region_t particle_rg = MM_MAKE_REGION();
  particle ** particles = generate_particles(n_particles, particle_rg);
  if (dump_p) dump_particles(particles, n_particles);
  simulate_steps(particles, n_particles, n_steps, dt);
  if (dump_p) dump_particles(particles, n_particles);
  MM_RESET_REGION(particle_rg);
#if UNDERLYING_MM_IS_SGC
  printf ("%d GCs\n", GC_gc_no);
#endif
  printf("nbody_main end\n");
}

#if BLOCK_TEST
extern int block_counter;
extern int block_frequency;
#endif
int st_main (int argc, char ** argv)
{
#if BLOCK_TEST
  if (argc != 4) {
    printf("usage: nbody n_particles n_steps block_frequency\n");
    st_wg_exit((void *)1);
  }
#else
  if (argc != 3) {
    printf ("usage: nbody n_particles n_steps\n");
    st_wg_exit((void *)1);
  }
#endif
  int n_particles = atoi(argv[1]);
  int n_steps = atoi(argv[2]);
  int dump_p = 0;
  t_real dt = 0.1;
#if BLOCK_TEST
  block_frequency = atoi(argv[3]);
  block_counter = block_frequency;
#endif
  
  MM_SETUP(/* GC_free_space_divisor = */2);
  printf("nbody_main %d %d\n", n_particles, n_steps);
  long t0 = st_current_time_ms();
  nbody_main(n_particles, n_steps, dt, dump_p);
  long t1 = st_current_time_ms();
  printf("TOTAL %d msec\n", t1 - t0);
  return 0;
}

