/*
 * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, December 18, 1998 */

/*
 * Declarations of mark stack.  Needed by marker and client supplied mark
 * routines.  To be included after gc_priv.h.
 */
#ifndef GC_MARK_H
# define GC_MARK_H

# define CUTBYTES_TO_PUSH 512 /* 512 */
# define CUTWORDS_TO_PUSH (BYTES_TO_WORDS(CUTBYTES_TO_PUSH))

# if !defined DO_ATOMIC_MARK
#   define SETMB_KIND 0
# elif DO_OPT_MOS == 1
#   define SETMB_KIND 3 /* 3 ... Mark with Opt Sync (Test-Compare&swap) */
# else
#   define SETMB_KIND 1 /* 3 ... Mark with Opt Sync (Test-Compare&swap) */
# endif

/* A client supplied mark procedure.  Returns new mark stack pointer.	*/
/* Primary effect should be to push new entries on the mark stack.	*/
/* Mark stack pointer values are passed and returned explicitly.	*/
/* Global variables decribing mark stack are not necessarily valid.	*/
/* (This usually saves a few cycles by keeping things in registers.)	*/
/* Assumed to scan about PROC_BYTES on average.  If it needs to do	*/
/* much more work than that, it should do it in smaller pieces by	*/
/* pushing itself back on the mark stack.				*/
/* Note that it should always do some work (defined as marking some	*/
/* objects) before pushing more than one entry on the mark stack.	*/
/* This is required to ensure termination in the event of mark stack	*/
/* overflows.								*/
/* This procedure is always called with at least one empty entry on the */
/* mark stack.								*/
/* Currently we require that mark procedures look for pointers in a	*/
/* subset of the places the conservative marker would.  It must be safe	*/
/* to invoke the normal mark procedure instead.				*/
# define PROC_BYTES 100
typedef struct ms_entry mse;
typedef struct ms_entry * (*mark_proc) GC_PROTO((SGC_MYENV_ARG_DECL1 word * addr, mse *mark_stack_ptr, mse *mark_stack_limit, word env));
						
# define LOG_MAX_MARK_PROCS 6
# define MAX_MARK_PROCS (1 << LOG_MAX_MARK_PROCS)
extern mark_proc GC_mark_procs[MAX_MARK_PROCS];
extern word GC_n_mark_procs;

/* Object descriptors on mark stack or in objects.  Low order two	*/
/* bits are tags distinguishing among the following 4 possibilities	*/
/* for the high order 30 bits.						*/
#define DS_TAG_BITS 2
#define DS_TAGS   ((1 << DS_TAG_BITS) - 1)
#define DS_LENGTH 0	/* The entire word is a length in bytes that	*/
			/* must be a multiple of 4.			*/
#define DS_BITMAP 1	/* 30 bits are a bitmap describing pointer	*/
			/* fields.  The msb is 1 iff the first word	*/
			/* is a pointer.				*/
			/* (This unconventional ordering sometimes	*/
			/* makes the marker slightly faster.)		*/
			/* Zeroes indicate definite nonpointers.  Ones	*/
			/* indicate possible pointers.			*/
			/* Only usable if pointers are word aligned.	*/
#   define BITMAP_BITS (WORDSZ - DS_TAG_BITS)
#define DS_PROC   2
			/* The objects referenced by this object can be */
			/* pushed on the mark stack by invoking		*/
			/* PROC(descr).  ENV(descr) is passed as the	*/
			/* last argument.				*/
#   define PROC(descr) \
		(GC_mark_procs[((descr) >> DS_TAG_BITS) & (MAX_MARK_PROCS-1)])
#   define ENV(descr) \
		((descr) >> (DS_TAG_BITS + LOG_MAX_MARK_PROCS))
#   define MAX_ENV \
  	      (((word)1 << (WORDSZ - DS_TAG_BITS - LOG_MAX_MARK_PROCS)) - 1)
#   define MAKE_PROC(proc_index, env) \
	    (((((env) << LOG_MAX_MARK_PROCS) | (proc_index)) << DS_TAG_BITS) \
	    | DS_PROC)
#define DS_PER_OBJECT 3	/* The real descriptor is at the		*/
			/* byte displacement from the beginning of the	*/
			/* object given by descr & ~DS_TAGS		*/
			
struct ms_entry {
    word * mse_start;   /* First word of object */
    word mse_descr;	/* Descriptor; low order two bits are tags,	*/
    			/* identifying the upper 30 bits as one of the	*/
    			/* following:					*/
};

#  define COPY_MSE(to_msep, from_msep) \
  (to_msep)->mse_start = (from_msep)->mse_start; \
  (to_msep)->mse_descr = (from_msep)->mse_descr

word GC_find_start GC_PROTO((register word current, register hdr *hhdr)); 

mse *GC_signal_mark_stack_overflow GC_PROTO((SGC_MYENV_ARG_DECL1 mse *msp)); 

# ifdef GATHERSTATS
#   define ADD_TO_ATOMIC(sz) GC_myg_gc_stat.atomic_in_use += (sz)
#   define ADD_TO_COMPOSITE(sz) GC_myg_gc_stat.composite_in_use += (sz)
# else
#   define ADD_TO_ATOMIC(sz)
#   define ADD_TO_COMPOSITE(sz)
# endif


#include <stdio.h>

#ifdef COUNT_CROWD /* don't mind! please ignore. */
#  define CCHDR(hhdr) \
if (ccprevhdr == hhdr) { \
    if (ccon == 0) { \
	PSM_counter_add(&hhdr->crowd_count, 1); \
	ccon = 1; \
	if ((GC_signed_word)hhdr->crowd_count > (GC_signed_word)hhdr->crowd_max) {\
	    hhdr->crowd_max = hhdr->crowd_count; \
        } \
    } \
} \
else { \
    if (ccon == 1) { \
	PSM_counter_add(&ccprevhdr->crowd_count, -1); \
	ccon = 0; \
    } \
    ccprevhdr = hhdr; \
}
#else
#  define CCHDR(hhdr)
#endif

/* Push the object obj with corresponding heap block header hhdr onto 	*/
/* the mark stack.							*/

#if DO_OPT_SLO == 1
# define PUSH_OBJ(obj, hhdr, mark_stack_top, mark_stack_limit) \
{ \
    register word _descr = (hhdr) -> hb_descr; \
        \
    DEMOWIN_MB_UPDATE(obj, hhdr); \
    if (_descr == 0) { \
    	ADD_TO_ATOMIC((hhdr) -> hb_sz); \
    } else { \
        ADD_TO_COMPOSITE((hhdr) -> hb_sz); \
	if ((hhdr) -> hb_sz >= CUTWORDS_TO_PUSH &&\
            (_descr & DS_TAGS) == DS_LENGTH) { \
          /* Split long object, Added by Endo */ \
          word sz = WORDS_TO_BYTES((hhdr) -> hb_sz); \
	  ptr_t p = (ptr_t)obj; \
	  word pushsz; \
          while (sz > 0) { \
            if (mark_stack_top+1 >= mark_stack_limit) { \
              mark_stack_top = GC_signal_mark_stack_overflow(SGC_MYENV_ARG1 mark_stack_top); \
              mark_stack_limit = GC_my_mark_stack.buf+GC_my_mark_stack.size; \
            } \
            pushsz = (sz >= CUTBYTES_TO_PUSH)? CUTBYTES_TO_PUSH: sz; \
            (mark_stack_top+1) -> mse_start = (word *)p; \
            (mark_stack_top+1) -> mse_descr = (((pushsz+DS_TAGS) & ~DS_TAGS) | DS_LENGTH); \
            GC_MSLOG_PUT(GC_MSC_PUSH, (word *)p);\
	    mark_stack_top++; \
            p += pushsz; \
            sz -= pushsz; \
	  } \
	} \
	else { \
          if (mark_stack_top+1 >= mark_stack_limit) { \
            mark_stack_top = GC_signal_mark_stack_overflow(SGC_MYENV_ARG1 mark_stack_top); \
            mark_stack_limit = GC_my_mark_stack.buf+GC_my_mark_stack.size; \
          } \
          (mark_stack_top+1) -> mse_start = (obj); \
          (mark_stack_top+1) -> mse_descr = _descr; \
          GC_MSLOG_PUT(GC_MSC_PUSH, (word *)obj);\
          mark_stack_top++; \
      } \
    } \
}
#else
# define PUSH_OBJ(obj, hhdr, mark_stack_top, mark_stack_limit) \
{ \
    register word _descr = (hhdr) -> hb_descr; \
        \
    DEMOWIN_MB_UPDATE(obj, hhdr); \
    if (_descr == 0) { \
    	ADD_TO_ATOMIC((hhdr) -> hb_sz); \
    } else { \
        ADD_TO_COMPOSITE((hhdr) -> hb_sz); \
        if (mark_stack_top+1 >= mark_stack_limit) { \
          mark_stack_top = GC_signal_mark_stack_overflow(SGC_MYENV_ARG1 mark_stack_top); \
          mark_stack_limit = GC_my_mark_stack.buf+GC_my_mark_stack.size; \
        } \
        (mark_stack_top+1) -> mse_start = (obj); \
        (mark_stack_top+1) -> mse_descr = _descr; \
        GC_MSLOG_PUT(GC_MSC_PUSH, (word *)obj);\
        mark_stack_top++; \
    } \
}
#endif

# if SETMB_KIND == 3  /* TCS: Test-Compare&Swap */

#define TEST_AND_MARK(current, hhdr, displ, mark_stack_top, mark_stack_limit) \
    { \
        register word volatile * mark_word_addr = hhdr -> hb_marks + divWORDSZ(displ); \
        register word mark_bit = (word)1 << modWORDSZ(displ); \
	if (PSM_test_and_set((unsigned long *)mark_word_addr, mark_bit)) { \
	    /* Succeeded */ \
	    CCHDR(hhdr); \
            PUSH_OBJ(((word *)(HBLKPTR(current)) + displ), hhdr, \
    	        mark_stack_top, mark_stack_limit) \
            n_marked ++; /* Added by Endo */ \
	} \
	else { \
	    /* Mark bit is already set */ \
	    n_hasmarked ++; /* Added by Endo */ \
	} \
    }

#elif SETMB_KIND == 1  /* With lock */

#define TEST_AND_MARK(current, hhdr, displ, mark_stack_top, mark_stack_limit) \
    { \
        register word * mark_word_addr = hhdr -> hb_marks + divWORDSZ(displ); \
        register word mark_word; \
        register word mark_bit = (word)1 << modWORDSZ(displ); \
          \
    	LOCK_MB(hhdr, divWORDSZ(displ));  /* Added by Endo */ \
        mark_word = *mark_word_addr; \
        if (!(mark_word & mark_bit)) { \
	    /* Mark bit is already set */ \
            *mark_word_addr = mark_word | mark_bit; \
	    UNLOCK_MB(hhdr, divWORDSZ(displ)); /* Added by Endo */ \
            PUSH_OBJ(((word *)(HBLKPTR(current)) + displ), hhdr, \
	         mark_stack_top, mark_stack_limit); \
            n_marked ++; /* Added by Endo */ \
        } \
	else { \
	    /* Mark bit is already set */ \
	    UNLOCK_MB(hhdr, divWORDSZ(displ)); /* Added by Endo */ \
	    n_hasmarked ++; /* Added by Endo */ \
        } \
    }

#elif SETMB_KIND == 2  /* Test-Test&Set */

#define TEST_AND_MARK(current, hhdr, displ, mark_stack_top, mark_stack_limit) \
    { \
        register word * mark_word_addr = hhdr -> hb_marks + divWORDSZ(displ); \
        register word mark_word; \
        register word mark_bit = (word)1 << modWORDSZ(displ); \
          \
        mark_word = *mark_word_addr; \
        if (mark_word & mark_bit) { \
	      /* Mark bit is already set */ \
	      n_hasmarked ++; /* Added by Endo */ \
        } \
	else { \
    	  LOCK_MB(hhdr, divWORDSZ(displ));  /* Added by Endo */ \
          mark_word = *mark_word_addr; \
          if (!(mark_word & mark_bit)) { \
	      /* Mark bit is already set */ \
              *mark_word_addr = mark_word | mark_bit; \
	      UNLOCK_MB(hhdr, divWORDSZ(displ)); /* Added by Endo */ \
              PUSH_OBJ(((word *)(HBLKPTR(current)) + displ), hhdr, \
	           mark_stack_top, mark_stack_limit); \
              n_marked ++; /* Added by Endo */ \
          } \
	  else { \
	      /* Mark bit is already set */ \
	      UNLOCK_MB(hhdr, divWORDSZ(displ)); /* Added by Endo */ \
	      n_hasmarked ++; /* Added by Endo */ \
          } \
	} \
    }

#elif SETMB_KIND == 0  /* No lock. Sequential only!! */

#define TEST_AND_MARK(current, hhdr, displ, mark_stack_top, mark_stack_limit) \
    { \
        register word * mark_word_addr = hhdr -> hb_marks + divWORDSZ(displ); \
        register word mark_word; \
        register word mark_bit = (word)1 << modWORDSZ(displ); \
          \
        mark_word = *mark_word_addr; \
        if (!(mark_word & mark_bit)) { \
	    /* Mark bit is already set */ \
            *mark_word_addr = mark_word | mark_bit; \
            PUSH_OBJ(((word *)(HBLKPTR(current)) + displ), hhdr, \
	         mark_stack_top, mark_stack_limit); \
            n_marked ++; /* Added by Endo */ \
        } \
	else { \
	    /* Mark bit is already set */ \
	    n_hasmarked ++; /* Added by Endo */ \
        } \
    }

#endif

#ifdef DO_REQ_REMOTE
# define REQ_REMOTE(p, hhdr) \
{ \
    int home = hhdr->home; \
    if (home != GC_my_pno && \
	SMALL_OBJ(WORDS_TO_BYTES((hhdr) -> hb_sz)) && \
	1/*GC_l_reqbuf(home).validsize == 0*/) { \
	int top; \
	struct reqbuf *prb = &GC_l_reqbuf(home);\
	PSM_mutex_lock(&prb->ml); \
	top = prb->validsize; \
	if (top >= prb->bufsize) { \
	    GC_err_printf0("[req_remote]: Overflow reqbuf...\n"); \
	} \
	else { \
	    prb->buf[top].obj = (word)p; \
	    prb->buf[top].hhdr = hhdr; \
	    prb->validsize = top+1; \
	} \
	PSM_mutex_unlock(&prb->ml); \
	continue; \
    } \
}
#else
# define REQ_REMOTE(p, hhdr) 
#endif

/* Push the contenst of current onto the mark stack if it is a valid	*/
/* ptr to a currently unmarked object.  Mark it.			*/

# define PUSH_CONTENTS(current, mark_stack_top, mark_stack_limit) \
{ \
    register int displ;  /* Displacement in block; first bytes, then words */ \
    register hdr * hhdr; \
    register map_entry_type map_entry; \
    \
    GET_HDR(current,hhdr); \
    if (hhdr == 0) {\
      GC_ADD_TO_BLACK_LIST_NORMAL(current); continue;\
    }\
    if (IS_FORWARDING_ADDR_OR_NIL(hhdr)) { \
      current = (word)HBLKPTR(current) + HDR_BYTES; /* From mark.c */\
      current = current - HBLKSIZE*(word)hhdr; \
      hhdr = HDR(current); \
    } \
    displ = HBLKDISPL(current); \
    map_entry = MAP_ENTRY((hhdr -> hb_map), displ); \
    if (map_entry == OBJ_INVALID) { \
        GC_ADD_TO_BLACK_LIST_NORMAL(current); continue; \
    } \
    displ = BYTES_TO_WORDS(displ); \
    displ -= map_entry; \
	\
    REQ_REMOTE(((word *)(HBLKPTR(current)) + displ), hhdr); \
    GC_CACHELOG_PUT_MBHDR(hhdr); \
    TEST_AND_MARK(current, hhdr, displ, mark_stack_top, mark_stack_limit); \
}

/*
 * Push a single value onto mark stack. Mark from the object pointed to by p.
 * GC_push_one is normally called by GC_push_regs, and thus must be defined.
 * P is considered valid even if it is an interior pointer.
 * Previously marked objects are not pushed.  Hence we make progress even
 * if the mark stack overflows.
 */

/* Modified by Endo. Assumes least_ha = GC_least_... */

# define GC_PUSH_ONE_STACK(p, least_ha, greatest_ha) \
    if ((ptr_t)(p) >= least_ha 	\
	 && (ptr_t)(p) < greatest_ha) {	\
	 GC_push_one_checked(SGC_MYENV_ARG1 p, TRUE);	\
    }

/* Put p into rootcopy. */
# define GC_EMIT_ONE_STACK(p, least_ha, greatest_ha) \
    if ((ptr_t)(p) >= least_ha 	\
	 && (ptr_t)(p) < greatest_ha) {	\
	 GC_emit_one_checked(SGC_MYENV_ARG1 p);	\
    }


/*
 * As above, but interior pointer recognition as for
 * normal for heap pointers.
 */
# ifdef ALL_INTERIOR_POINTERS
#   define AIP TRUE
# else
#   define AIP FALSE
# endif

/* Modified by Endo. Assumes least_ha = GC_least_... */

# define GC_PUSH_ONE_HEAP(p, least_ha, greatest_ha) \
    if ((ptr_t)(p) >= least_ha 	\
	 && (ptr_t)(p) < greatest_ha) {	\
	 GC_push_one_checked(SGC_MYENV_ARG1 p,AIP);	\
    }


typedef int mark_state_t;	/* Current state of marking, as follows:*/
				/* Used to remember where we are during */
				/* concurrent marking.			*/

				/* We say something is dirty if it was	*/
				/* written since the last time we	*/
				/* retrieved dirty bits.  We say it's 	*/
				/* grungy if it was marked dirty in the	*/
				/* last set of bits we retrieved.	*/
				
				/* Invariant I: all roots and marked	*/
				/* objects p are either dirty, or point */
				/* objects q that are either marked or	*/
				/* a pointer to q appears in a range	*/
				/* on the mark stack.			*/

# define MS_NONE 0		/* No marking in progress. I holds.	*/
				/* Mark stack is empty.			*/

# define MS_PUSH_RESCUERS 1	/* Rescuing objects are currently 	*/
				/* being pushed.  I holds, except	*/
				/* that grungy roots may point to 	*/
				/* unmarked objects, as may marked	*/
				/* grungy objects above scan_ptr.	*/

# define MS_PUSH_UNCOLLECTABLE 2
				/* I holds, except that marked 		*/
				/* uncollectable objects above scan_ptr */
				/* may point to unmarked objects.	*/
				/* Roots may point to unmarked objects	*/

# define MS_ROOTS_PUSHED 3	/* I holds, mark stack may be nonempty  */

# define MS_PARTIALLY_INVALID 4	/* I may not hold, e.g. because of M.S. */
				/* overflow.  However marked heap	*/
				/* objects below scan_ptr point to	*/
				/* marked or stacked objects.		*/

# define MS_INVALID 5		/* I may not hold.			*/

#endif  /* GC_MARK_H */
