/*
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1995 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, June 3, 1999 */

# include <stdio.h>
# include "gc_priv.h"
/* # include "gc_mark.h" */

/* We put this here to minimize the risk of inlining. */
/*VARARGS*/
void GC_noop GC_PROTO((void *p, ...)) {}

/* Single argument version, robust against whole program analysis. */
void GC_noop1(x)
word x;
{
    static VOLATILE word sink;

    sink = x;
}

static void clear_marks_for_block GC_PROTO((struct hblk *h, word dummy)); 

mark_proc GC_mark_procs[MAX_MARK_PROCS] = {0};
word GC_n_mark_procs = 0;

/* Initialize GC_obj_kinds properly and standard free lists properly.  	*/
/* This must be done statically since they may be accessed before 	*/
/* GC_init is called.							*/
/* It's done here, since we need to deal with mark descriptors.		*/

# ifdef ATOMIC_UNCOLLECTABLE
#   ifdef STUBBORN_ALLOC
      int GC_n_kinds = 5;
#   else
      int GC_n_kinds = 4;
#   endif
# else
#   ifdef STUBBORN_ALLOC
      int GC_n_kinds = 4;
#   else
      int GC_n_kinds = 3;
#   endif
# endif

/* Added by Endo. */
int GC_init_obj_kinds()
{
    /* PTRFREE */
    GC_obj_kinds[PTRFREE].ok_descriptor = 0 | DS_LENGTH;
    GC_obj_kinds[PTRFREE].ok_relocate_descr = FALSE;
    GC_obj_kinds[PTRFREE].ok_init = FALSE;

    /* NORMAL  */
#if defined(ADD_BYTE_AT_END) && ALIGNMENT > DS_TAGS
    GC_obj_kinds[NORMAL].ok_descriptor = (word)(-ALIGNMENT) | DS_LENGTH;
#else
    GC_obj_kinds[NORMAL].ok_descriptor = 0 | DS_LENGTH;
#endif
    GC_obj_kinds[NORMAL].ok_relocate_descr = TRUE /* add length to descr */;
    GC_obj_kinds[NORMAL].ok_init = TRUE;

    /* UNCOLLECTABLE */
    GC_obj_kinds[UNCOLLECTABLE].ok_descriptor = 0 | DS_LENGTH;
    GC_obj_kinds[UNCOLLECTABLE].ok_relocate_descr = TRUE /* add length to descr */;
    GC_obj_kinds[UNCOLLECTABLE].ok_init = TRUE;

#ifdef ATOMIC_UNCOLLECTABLE
    /*AUNCOLLECTABLE*/
    GC_obj_kinds[AUNCOLLECTABLE].ok_descriptor = 0 | DS_LENGTH;
    GC_obj_kinds[AUNCOLLECTABLE].ok_relocate_descr = FALSE /* add length to descr */;
    GC_obj_kinds[AUNCOLLECTABLE].ok_init = FALSE;
#endif

#ifdef STUBBORN_ALLOC
    /*STUBBORN*/
    GC_obj_kinds[STUBBORN].ok_descriptor = 0 | DS_LENGTH;
    GC_obj_kinds[STUBBORN].ok_relocate_descr = TRUE /* add length to descr */;
    GC_obj_kinds[STUBBORN].ok_init = TRUE;
#endif
    return 0;
}

# define INITIAL_MARK_STACK_SIZE (4*HBLKSIZE)
		/* INITIAL_MARK_STACK_SIZE * sizeof(mse) should be a 	*/
		/* multiple of HBLKSIZE.				*/

#define INITIAL_REQBUF_SIZE (16*HBLKSIZE)

/*
 * Limits of stack for GC_mark routine.
 * All ranges between GC_mark_stack(incl.) and GC_mark_stack_top(incl.) still
 * need to be marked from.
 */

word GC_n_rescuing_pages;	/* Number of dirty pages we marked from */
				/* excludes ptrfree pages, etc.		*/

GC_bool GC_collection_in_progress(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    return(GC_my_mark_state != MS_NONE);
}

/* clear all mark bits in the header */
void GC_clear_hdr_marks(hhdr)
register hdr * hhdr;
{
    BZERO(hhdr -> hb_marks, MARK_BITS_SZ*sizeof(word));
    BZERO(hhdr -> mutex, MARK_BITS_SZ*sizeof(PSM_mutex_t));
}

/* Set all mark bits in the header.  Used for uncollectable blocks. */
void GC_set_hdr_marks(hhdr)
register hdr * hhdr;
{
    register int i;

    for (i = 0; i < MARK_BITS_SZ; ++i) {
    	hhdr -> hb_marks[i] = ONES;
    }
}

/*
 * Clear all mark bits associated with block h.
 */
/*ARGSUSED*/
static void clear_marks_for_block(h, dummy)
struct hblk *h;
word dummy;
{
    register hdr * hhdr = HDR(h);

#ifdef COUNT_CROWD
    hhdr->crowd_count = 0;
    hhdr->crowd_max = 0;
#endif
    
    if (IS_UNCOLLECTABLE(hhdr -> hb_obj_kind)) return;
        /* Mark bit for these is cleared only once the object is 	*/
        /* explicitly deallocated.  This either frees the block, or	*/
        /* the bit is cleared once the object is on the free list.	*/
    GC_clear_hdr_marks(hhdr);
}

/* Slow but general routines for setting/clearing/asking about mark bits */
void GC_set_mark_bit(p)
ptr_t p;
{
    register struct hblk *h = HBLKPTR(p);
    register hdr * hhdr = HDR(h);
    register int word_no = (word *)p - (word *)h;
    
    set_mark_bit_from_hdr(hhdr, word_no);
}

void GC_clear_mark_bit(p)
ptr_t p;
{
    register struct hblk *h = HBLKPTR(p);
    register hdr * hhdr = HDR(h);
    register int word_no = (word *)p - (word *)h;
    
    clear_mark_bit_from_hdr(hhdr, word_no);
}

GC_bool GC_is_marked(p)
ptr_t p;
{
    register struct hblk *h = HBLKPTR(p);
    register hdr * hhdr = HDR(h);
    register int word_no = (word *)p - (word *)h;
    
    return(mark_bit_from_hdr(hhdr, word_no));
}


/*
 * Clear mark bits in all allocated heap blocks.  This invalidates
 * the marker invariant, and sets GC_mark_state to reflect this.
 * (This implicitly starts marking to reestablish the
 */
void GC_clear_marks(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_apply_to_all_blocks(clear_marks_for_block, (word)0);
    GC_my_objects_are_marked = FALSE;
    GC_my_mark_state = MS_INVALID;
    GC_scan_ptr = 0;
    GC_scan_ptr_flag = 0;
}

void GC_mark_stack_expand GC_PROTO((SGC_MYENV_ARG_DECL1 word n)); 

#ifdef DO_REQ_REMOTE
#define REMAIN_WORKS_ON_MS() \
  (GC_my_mark_stack.top >= GC_my_mark_stack.bottom || \
   GC_myg_reqbuf.validsize > 0)
#else
#define REMAIN_WORKS_ON_MS() \
  (GC_my_mark_stack.top >= GC_my_mark_stack.bottom)
#endif

#define PUSH_LOCALROOTS_BEFORE_UNCOLLECTABLE 1

/* Perform a small amount of marking.			*/
/* We try to touch roughly a page of memory.		*/
/* Return TRUE if we just finished a mark phase.	*/
GC_bool GC_mark_some(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    switch(GC_my_mark_state) {
    	case MS_NONE:
    	    return(FALSE);
    	    
    	case MS_PUSH_RESCUERS:
	    /* TENUKI. incremental GC is not supported. */
    	    return(FALSE);
    	
    	case MS_PUSH_UNCOLLECTABLE:
    	    if (GC_my_mark_stack.top
    	        >= GC_my_mark_stack.bottom + INITIAL_MARK_STACK_SIZE/4) {
		GC_LDLOG_PUT(GC_LDC_S, GC_LDK_MFMS);
    	        GC_mark_from_mark_stack(SGC_MYENV_ARG0);
		GC_LDLOG_PUT(GC_LDC_E, GC_LDK_MFMS);
    	        return(FALSE);
    	    }
	    else if (PSM_mutex_trylock(&GC_scan_ptr_ml) == 0) {
		GC_LDLOG_PUT(GC_LDC_S, GC_LDK_MFMS);
    	        GC_mark_from_mark_stack(SGC_MYENV_ARG0);
		GC_LDLOG_PUT(GC_LDC_E, GC_LDK_MFMS);
    	        return(FALSE);
	    }
	    else {
		if (GC_scan_ptr_flag == 0) {
		    GC_LDLOG_PUT(GC_LDC_S, GC_LDK_PROOT);
		    GC_scan_ptr =
			GC_push_next_marked_uncollectable(SGC_MYENV_ARG1 GC_scan_ptr);
		    GC_LDLOG_PUT(GC_LDC_E, GC_LDK_PROOT);
		    if (GC_scan_ptr == 0) {
			/* Heap scan is finished. */
			GC_scan_ptr_flag = 1;
		    }
		}
		PSM_mutex_unlock(&GC_scan_ptr_ml);
    	        if (GC_scan_ptr_flag == 1) {
#ifndef PUSH_LOCALROOTS_BEFORE_UNCOLLECTABLE
		    GC_push_roots(SGC_MYENV_ARG0);
#endif
    	    	    GC_my_objects_are_marked = TRUE;
    	    	    if (GC_my_mark_state != MS_INVALID) {
    	    	        GC_my_mark_state = MS_ROOTS_PUSHED;
    	    	    }
    	    	}
    	    }
    	    return(FALSE);
    	
    	case MS_ROOTS_PUSHED:
	    if (REMAIN_WORKS_ON_MS()) {
		GC_LDLOG_PUT(GC_LDC_S, GC_LDK_MFMS);
    	        GC_mark_from_mark_stack(SGC_MYENV_ARG0);
		GC_LDLOG_PUT(GC_LDC_E, GC_LDK_MFMS);
		if (!REMAIN_WORKS_ON_MS()) {
		    GC_my_mark_state = MS_NONE;
		    return TRUE;
		}
    	        return(FALSE);
    	    } else {
    	        GC_my_mark_state = MS_NONE;
    	        if (GC_my_mark_stack_too_small) {
    	            GC_mark_stack_expand(SGC_MYENV_ARG1 2*GC_my_mark_stack.size);
    	        }
    	        return(TRUE);
    	    }
    	    
    	case MS_INVALID:
    	case MS_PARTIALLY_INVALID:
	    if (!GC_my_objects_are_marked) {
		GC_my_mark_state = MS_PUSH_UNCOLLECTABLE;
#ifdef PUSH_LOCALROOTS_BEFORE_UNCOLLECTABLE
		GC_push_roots(SGC_MYENV_ARG0);
#endif
		return(FALSE);
	    }
	    fprintf(stderr, "WARNING: bad state in GC_mark_some(SGC_MYENV_ARG0)?\n");
	    return FALSE;
#if 0 /* Now partially invalid ms is disabled */
	    if (REMAIN_WORKS_ON_MS()) {
		GC_LDLOG_PUT(GC_LDC_S, GC_LDK_MFMS);
    	        GC_mark_from_mark_stack(SGC_MYENV_ARG0);
		GC_LDLOG_PUT(GC_LDC_E, GC_LDK_MFMS);
    	        return(FALSE);
    	    }
    	    if (GC_scan_ptr == 0
    	        && (GC_my_mark_state == MS_INVALID || GC_my_mark_stack_too_small)) {
		GC_mark_stack_expand(SGC_MYENV_ARG1 2*GC_my_mark_stack.size);
		GC_my_mark_state = MS_PARTIALLY_INVALID;
    	    }
    	    GC_scan_ptr = GC_push_next_marked(SGC_MYENV_ARG1 GC_scan_ptr);
    	    if (GC_scan_ptr == 0 && GC_my_mark_state == MS_PARTIALLY_INVALID) {
    	    	GC_push_roots(SGC_MYENV_ARG0);
    	    	GC_my_objects_are_marked = TRUE;
    	    	if (GC_my_mark_state != MS_INVALID) {
    	    	    GC_my_mark_state = MS_ROOTS_PUSHED;
    	    	}
    	    }
    	    return(FALSE);
#endif
    	default:
    	    ABORT("GC_mark_some: bad state");
    	    return(FALSE);
    }
}


GC_bool GC_mark_stack_empty(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    return(GC_my_mark_stack.top < GC_my_mark_stack.bottom);
}	

#ifdef PROF_MARKER
    word GC_prof_array[10];
#   define PROF(n) GC_prof_array[n]++
#else
#   define PROF(n)
#endif

/* Given a pointer to someplace other than a small object page or the	*/
/* first page of a large object, return a pointer either to the		*/
/* start of the large object or NIL.					*/
/* In the latter case black list the address current.			*/
/* Returns NIL without black listing if current points to a block	*/
/* with IGNORE_OFF_PAGE set.						*/
/*ARGSUSED*/
word GC_find_start(current, hhdr)
register word current;
register hdr * hhdr;
{
#   ifdef ALL_INTERIOR_POINTERS
	if (hhdr != 0) {
	    register word orig = current;
	    
	    current = (word)HBLKPTR(current) + HDR_BYTES;
	    do {
	      current = current - HBLKSIZE*(word)hhdr;
	      hhdr = HDR(current);
	    } while(IS_FORWARDING_ADDR_OR_NIL(hhdr));
	    /* current points to the start of the large object */
	    if (hhdr -> hb_flags & IGNORE_OFF_PAGE) return(0);
	    if ((word *)orig - (word *)current
	         >= (ptrdiff_t)(hhdr->hb_sz)) {
	        /* Pointer past the end of the block */
	        GC_ADD_TO_BLACK_LIST_NORMAL(orig);
	        return(0);
	    }
	    return(current);
	} else {
	    GC_ADD_TO_BLACK_LIST_NORMAL(current);
	    return(0);
        }
#   else
        GC_ADD_TO_BLACK_LIST_NORMAL(current);
        return(0);
#   endif
}

void GC_invalidate_mark_state(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_my_mark_state = MS_INVALID;
    GC_my_mark_stack.bottom = GC_my_mark_stack.buf;
    GC_my_mark_stack.watermark = GC_my_mark_stack.buf;
    GC_my_mark_stack.top = GC_my_mark_stack.bottom-1;
    PSM_mutex_init(&GC_my_mark_stack.ml);
}

mse * GC_signal_mark_stack_overflow(SGC_MYENV_ARG1 msp)
SGC_MYENV_ARG_DECL2
mse * msp;
{
    DECLARE_MYENV
#if 1 /* Modified by Endo */
    GC_my_mark_stack.top = msp;
    GC_mark_stack_expand(SGC_MYENV_ARG1 GC_my_mark_stack.size * 2);
    return GC_my_mark_stack.top;
#else
    GC_my_mark_state = MS_INVALID;
#   ifdef PRINTSTATS
	GC_printf1("Mark stack overflow; current size = %lu entries\n",
	    	    GC_my_mark_stack.size);
#    endif
     return(msp-INITIAL_MARK_STACK_SIZE/8);
#endif
}

# define SPLIT_RANGE_WORDS 512  /* Must be power of 2.	*/
# define INITIAL_CREDIT (1*HBLKSIZE)

/*
 * Mark objects pointed to by the regions described by
 * mark stack entries between GC_mark_stack and GC_mark_stack_top,
 * inclusive.  Assumes the upper limit of a mark stack entry
 * is never 0.  A mark stack entry never has size 0.
 * We try to traverse on the order of a hblk of memory before we return.
 * Caller is responsible for calling this until the mark stack is empty.
 */
void GC_mark_from_mark_stack(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
#if DO_LOAD_DIST == 4
  mse * volatile * p_mark_stack_bottom = &GC_my_mark_stack.bottom;
  mse * volatile * p_mark_stack_top = &GC_my_mark_stack.top;
  PSM_mutex_t *p_mark_stack_ml = &GC_my_mark_stack.ml;
#elif DO_LOAD_DIST == 5
  mse * mark_stack_watermark_reg = GC_my_mark_stack.watermark;
  mse * volatile * p_mark_stack_watermark = &GC_my_mark_stack.watermark;
  PSM_mutex_t *p_mark_stack_ml = &GC_my_mark_stack.ml;
  mse * volatile * p_mark_stack_bottom = &GC_my_mark_stack.bottom;
#else
  mse * mark_stack_bottom_reg = GC_my_mark_stack.bottom;
#endif
  mse * mark_stack_top_reg = GC_my_mark_stack.top;
  mse * mark_stack_limit = &(GC_my_mark_stack.buf[GC_my_mark_stack.size]);
  int credit = INITIAL_CREDIT;/* Remaining credit for marking work	*/
  register word * current_p;	/* Pointer to current candidate ptr.	*/
  register word current;	/* Candidate pointer.			*/
  register word * limit = 0;	/* (Incl) limit of current candidate 	*/
  				/* range				*/
  register word descr;
  register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
  register ptr_t least_ha = GC_least_plausible_heap_addr;
  word n_marked; /* Added by Endo for measurement */
  word n_hasmarked; /* Added by Endo for measurement */
  word words_scaned; /* Added by Endo for measurement */
#ifdef DO_PICLOG
  int pe;
#endif
#ifdef COUNT_CROWD
  hdr * ccprevhdr = NULL;
  int ccon = 0;
#endif

#ifdef DO_LDLOG
  int is_longobject = 0;
  int is_root = 0;
#endif

  GC_my_objects_are_marked = TRUE;

#ifdef DO_REQ_REMOTE
  GC_mark_from_reqbuf(SGC_MYENV_ARG0);
#if DO_LOAD_DIST == 4
#elif DO_LOAD_DIST == 5
  mark_stack_watermark_reg = GC_my_mark_stack.watermark;
  mark_stack_bottom_reg = GC_my_mark_stack.bottom;
#else
  mark_stack_bottom_reg = GC_my_mark_stack.bottom;
#endif
  mark_stack_top_reg = GC_my_mark_stack.top;
  mark_stack_limit = &(GC_my_mark_stack.buf[GC_my_mark_stack.size]);
#endif

  GC_PICLOG_CHANGE_MODE(1);

#if DO_LOAD_DIST == 4

  /* The bottom object may be stolen, */
  /* so empty-judgement requires lock. */
  while (1) {
      MUTEX_LOCK_LOG(p_mark_stack_ml, GC_LDK_WMQ);
      if (mark_stack_top_reg < *p_mark_stack_bottom || credit < 0) {
	  PSM_mutex_unlock(p_mark_stack_ml);
	  break;
      }

#elif DO_LOAD_DIST == 5

  while (credit >= 0) {
      if (mark_stack_top_reg < mark_stack_watermark_reg) {
	  int nmse;
	  MUTEX_LOCK_LOG(p_mark_stack_ml, GC_LDK_WMQ);
	  nmse = mark_stack_watermark_reg - (*p_mark_stack_bottom);
	  if (nmse <= 0) {
	      PSM_mutex_unlock(p_mark_stack_ml);
	      break;
	  } 
	  mark_stack_watermark_reg -= (nmse+1)/2;
	  *p_mark_stack_watermark = mark_stack_watermark_reg;
	  PSM_mutex_unlock(p_mark_stack_ml);
      }

#else /* DO_LOAD_DIST */

# ifdef OS2 /* Use untweaked version to circumvent compiler problem */
  while (mark_stack_top_reg >= mark_stack_bottom_reg && credit >= 0) {
# else
  while ((((ptr_t)mark_stack_top_reg - (ptr_t)mark_stack_bottom_reg) | credit)
  	>= 0) {
# endif

#endif /* DO_LOAD_DIST */

    current_p = mark_stack_top_reg -> mse_start;
    descr = mark_stack_top_reg -> mse_descr;

  retry:  
    if (descr & ((~(WORDS_TO_BYTES(SPLIT_RANGE_WORDS) - 1)) | DS_TAGS)) {
      word tag = descr & DS_TAGS;
      switch(tag) {
        case DS_LENGTH:
          /* Large length.					        */
          /* Process part of the range to avoid pushing too much on the	*/
          /* stack.							*/
          mark_stack_top_reg -> mse_start =
         	limit = current_p + SPLIT_RANGE_WORDS-1;
          mark_stack_top_reg -> mse_descr -=
          		WORDS_TO_BYTES(SPLIT_RANGE_WORDS-1);
          /* Make sure that pointers overlapping the two ranges are	*/
          /* considered. 						*/
          limit = (word *)((char *)limit + sizeof(word) - ALIGNMENT);
#ifdef DO_LDLOG
      is_longobject = 1;
#endif
          break;
        case DS_BITMAP:
          mark_stack_top_reg--;
          descr &= ~DS_TAGS;
          credit -= WORDS_TO_BYTES(WORDSZ/2); /* guess */
          while (descr != 0) {
            if ((signed_word)descr < 0) {
		n_marked = 0;
		n_hasmarked = 0;
              current = *current_p++;
              descr <<= 1;
              if ((ptr_t)current < least_ha) continue;
              if ((ptr_t)current >= greatest_ha) continue;
              PUSH_CONTENTS(current, mark_stack_top_reg, mark_stack_limit);
		GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MARKED, n_marked);
		GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_HASMARKED, n_hasmarked);
            } else {
              descr <<= 1;
              current_p++;
            }
          }
          continue;
        case DS_PROC:
          mark_stack_top_reg--;
          credit -= PROC_BYTES;
          mark_stack_top_reg =
              (*PROC(descr))
              	    (SGC_MYENV_ARG1 current_p, mark_stack_top_reg,
              	    mark_stack_limit, ENV(descr));
          continue;
        case DS_PER_OBJECT:
          descr = *(word *)((ptr_t)current_p + descr - tag);
          goto retry;
      }
    } else {
      mark_stack_top_reg--;
      GC_MSLOG_PUT(GC_MSC_POP, current_p);
      limit = (word *)(((ptr_t)current_p) + (word)descr);
#ifdef DO_LDLOG
      is_longobject = (descr >= CUTBYTES_TO_PUSH-8)? 1: 0;
#endif
    }

#if DO_LOAD_DIST == 4
    /* Update mark_stack_top and unlock */
    *p_mark_stack_top = mark_stack_top_reg;
    PSM_mutex_unlock(p_mark_stack_ml);
#elif DO_LOAD_DIST == 5
#else /* DO_LOAD_DIST */
#endif /* DO_LOAD_DIST */

#ifdef DO_MARKLOG
    n_marked = 0;
    n_hasmarked = 0;
    words_scaned = ((char *)limit - (char *)current_p) / ALIGNMENT + 1;
#endif

    /* The simple case in which we're scanning a range.	*/
    credit -= (ptr_t)limit - (ptr_t)current_p;
    limit -= 1;

    GC_CACHELOG_BEFORE_LOOP();
    while (current_p <= limit) {
	current = *current_p;
	current_p = (word *)((char *)current_p + ALIGNMENT);
	if ((ptr_t)current < least_ha) continue;
	if ((ptr_t)current >= greatest_ha) continue;
	PUSH_CONTENTS(current, mark_stack_top_reg, mark_stack_limit);
    }

    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MARKED, n_marked);
    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_HASMARKED, n_hasmarked);
    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_NONPTR, words_scaned - n_marked - n_hasmarked);
    GC_CACHELOG_PUT_OBJSIZE(words_scaned);
#ifdef USE_DEMOWIN
    GC_demowin_ms_update();
#endif
  }
  GC_my_mark_stack.top = mark_stack_top_reg;
  GC_PICLOG_CHANGE_MODE(0);

#ifdef USE_DEMOWIN
    GC_demowin_ms_update();
#endif
#ifdef COUNT_CROWD
    if (ccon == 1 && ccprevhdr) {
	PSM_counter_add(&ccprevhdr->crowd_count, -1);
	ccon = 0;
    }
#endif
}

#ifdef DO_REQ_REMOTE
/*******************************************************************/
/* GC_my_envp is not refered. */
int GC_reqbuf_init(GC_thread envp)
{
    struct reqbuf_entry *p;
    word size;

    size = INITIAL_REQBUF_SIZE;
    p = (struct reqbuf_entry *)
	GET_MEM(sizeof(struct reqbuf_entry)*size);
    if (p == NULL) {
	fprintf(stderr, "[GC_reqbuf_init] Cannot alloc buffer.\n");
	GC_system_terminate();
	exit(1);
    }
    GC_touch_memory_region((char *)p, sizeof(struct reqbuf_entry)*size);
    GC_lg_reqbuf(envp).buf = p;
    GC_lg_reqbuf(envp).bufsize = size;
    GC_lg_reqbuf(envp).validsize = 0;
    PSM_mutex_init(&GC_lg_reqbuf(envp).ml);

    return 0;
}

int GC_mark_from_reqbuf(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int i;
    int top;
    word current;
    hdr * hhdr;
    struct reqbuf_entry *p;
    struct reqbuf_entry *limit;
    mse * mark_stack_top_reg = GC_my_mark_stack.top;
    mse * mark_stack_limit = &(GC_my_mark_stack.buf[GC_my_mark_stack.size]);
    word n_marked = 0;
    word n_hasmarked = 0;
    word words_scaned = 0;

    MUTEX_LOCK_LOG(&GC_myg_reqbuf.ml, GC_LDK_WREQB);
    p = GC_myg_reqbuf.buf;
    limit = p+GC_myg_reqbuf.validsize;
    while (p < limit) {
	int displ;

	current = p->obj;
	hhdr = p->hhdr;
	p++;

	displ = BYTES_TO_WORDS(HBLKDISPL(current));
	TEST_AND_MARK(current, hhdr, displ,
		      mark_stack_top_reg, mark_stack_limit);
    }
    GC_myg_reqbuf.validsize = 0;
    PSM_mutex_unlock(&GC_myg_reqbuf.ml);

    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MARKED, n_marked);
    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_HASMARKED, n_hasmarked);
    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_NONPTR, words_scaned - n_marked - n_hasmarked);

    GC_my_mark_stack.top = mark_stack_top_reg;

    return 0;
}
#endif

/*******************************************************************/
/* GC_my_envp is not refered. */
int GC_mark_stack_init(GC_thread envp)
{
    mse *buf;
    word size;
    
#if 1 || defined SGC_DEBUGPRINT
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!MSINIT:0! envp broken %08x, pid %d\n",
		       envp, GC_l_pid(envp));
    }
#endif
    size = INITIAL_MARK_STACK_SIZE;
    buf = (mse *)GET_MEM(size*sizeof(struct ms_entry));
    if (buf == NULL) {
	fprintf(stderr, "[GC_mark_stack_init] Cannot alloc mark_stack.\n");
	GC_system_terminate();
	exit(1);
    }
#if 1 || defined SGC_DEBUGPRINT
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!MSINIT:A! envp broken %08x, pid %d\n",
		       envp, GC_l_pid(envp));
    }
#endif
    GC_touch_memory_region((char *)buf, size*sizeof(struct ms_entry));
    
    GC_l_mark_stack(envp).size = size;
    GC_l_mark_stack(envp).buf = buf;
    GC_l_mark_stack(envp).bottom = buf;
    GC_l_mark_stack(envp).watermark = buf;
    GC_l_mark_stack(envp).top = buf-1;
    GC_l_mark_stack_too_small(envp) = FALSE;
    return 0;
}

/* Allocate or reallocate space for mark stack of size s words  */
/* May silently fail.						*/
void GC_mark_stack_expand(SGC_MYENV_ARG1 n)
SGC_MYENV_ARG_DECL2
     word n;
{
    DECLARE_MYENV
    mse * new_stack = (mse *)GC_scratch_alloc(SGC_MYENV_ARG1 n * sizeof(struct ms_entry));

    GC_my_mark_stack_too_small = FALSE;

    if (new_stack != 0) {
	word displ = HBLKDISPL(GC_my_mark_stack.buf);
	word size = GC_my_mark_stack.size * sizeof(struct ms_entry);
	word nmse;

	PSM_mutex_lock(&GC_my_mark_stack.ml);

	/* Copy contents of old stack into new stack. */
	nmse = GC_my_mark_stack.top + 1 - GC_my_mark_stack.bottom;

	BCOPY(GC_my_mark_stack.bottom, new_stack,
	      nmse * sizeof(struct ms_entry));
	
	/* Recycle old space */
	if (displ == 0) {
	    GC_add_to_heap(SGC_MYENV_ARG1 (struct hblk *)GC_my_mark_stack.buf, size);
	} else {
	    GC_add_to_heap(SGC_MYENV_ARG1 (struct hblk *)
			   ((word)GC_my_mark_stack.buf - displ + HBLKSIZE),
			   size - HBLKSIZE);
	}
	GC_my_mark_stack.size = n;
	GC_my_mark_stack.buf = new_stack;
	GC_my_mark_stack.bottom = new_stack;
	GC_my_mark_stack.top = new_stack + nmse - 1;

	PSM_mutex_unlock(&GC_my_mark_stack.ml);

#	  ifdef PRINTSTATS
	GC_err_printf1("Grew mark stack to %lu frames\n",
		   (unsigned long) GC_my_mark_stack.size);
#	  endif
    } else {
#	  ifdef PRINTSTATS
	GC_err_printf1("Failed to grow mark stack to %lu frames\n",
		   (unsigned long) n);
#	  endif
    }
    return;
}

/*
 * Push all locations between b and t onto the mark stack.
 * b is the first location to be checked. t is one past the last
 * location to be checked.
 * Should only be used if there is no possibility of mark stack
 * overflow.
 */
void GC_push_all(SGC_MYENV_ARG1 bottom, top)
SGC_MYENV_ARG_DECL2
ptr_t bottom;
ptr_t top;
{
    DECLARE_MYENV
    register word length;

    bottom = (ptr_t)(((word) bottom + ALIGNMENT-1) & ~(ALIGNMENT-1));
    top = (ptr_t)(((word) top) & ~(ALIGNMENT-1));
    if (top == 0 || bottom == top) return;
    GC_my_mark_stack.top++;
    if (GC_my_mark_stack.top >=
	GC_my_mark_stack.buf + GC_my_mark_stack.size) {
	ABORT("unexpected mark stack overflow (GC_push_all)");
    }
    length = top - bottom;
#   if DS_TAGS > ALIGNMENT - 1
	length += DS_TAGS;
	length &= ~DS_TAGS;
#   endif
    GC_my_mark_stack.top -> mse_start = (word *)bottom;
    GC_my_mark_stack.top -> mse_descr = length;
}

/* Added by Endo */
/* Called by App/GC proc. */
/* This resembles GC_push_all, but this pushes each contained pointer */
/* rather than bottom-top region itself. */
void GC_push_all_contents(SGC_MYENV_ARG1 bottom, top)
SGC_MYENV_ARG_DECL2
ptr_t bottom;
ptr_t top;
{
    DECLARE_MYENV
/*    register word length; */
    register int n;
    register int limit;
    register ptr_t *p;
    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
    register ptr_t least_ha = GC_least_plausible_heap_addr;

    bottom = (ptr_t)(((word) bottom + ALIGNMENT-1) & ~(ALIGNMENT-1));
    top = (ptr_t)(((word) top) & ~(ALIGNMENT-1));
    if (top == 0 || bottom == top) return;

    for (p = (ptr_t *)bottom; p < (ptr_t *)top; p++) {
	GC_PUSH_ONE_STACK((word)*p, least_ha, greatest_ha);
    }
}

/*
 * Analogous to the above, but push only those pages that may have been
 * dirtied.  A block h is assumed dirty if dirty_fn(h) != 0.
 * We use push_fn to actually push the block.
 * Will not overflow mark stack if push_fn pushes a small fixed number
 * of entries.  (This is invoked only if push_fn pushes a single entry,
 * or if it marks each object before pushing it, thus ensuring progress
 * in the event of a stack overflow.)
 */
void GC_push_dirty(SGC_MYENV_ARG1 bottom, top, dirty_fn, push_fn)
SGC_MYENV_ARG_DECL2
ptr_t bottom;
ptr_t top;
int (*dirty_fn) GC_PROTO((struct hblk * h));
void (*push_fn) GC_PROTO((SGC_MYENV_ARG_DECL1 ptr_t bottom, ptr_t top));
{
    DECLARE_MYENV
    register struct hblk * h;

    bottom = (ptr_t)(((long) bottom + ALIGNMENT-1) & ~(ALIGNMENT-1));
    top = (ptr_t)(((long) top) & ~(ALIGNMENT-1));

    if (top == 0 || bottom == top) return;
    h = HBLKPTR(bottom + HBLKSIZE);
    if (top <= (ptr_t) h) {
  	if ((*dirty_fn)(h-1)) {
	    (*push_fn)(SGC_MYENV_ARG1 bottom, top);
	}
	return;
    }
    if ((*dirty_fn)(h-1)) {
        (*push_fn)(SGC_MYENV_ARG1 bottom, (ptr_t)h);
    }
    while ((ptr_t)(h+1) <= top) {
	if ((*dirty_fn)(h)) {
	    if ((word)(GC_my_mark_stack.top - GC_my_mark_stack.buf)
		> 3 * GC_my_mark_stack.size / 4) {
	 	/* Danger of mark stack overflow */
		(*push_fn)(SGC_MYENV_ARG1 (ptr_t)h, top);
		return;
	    } else {
		(*push_fn)(SGC_MYENV_ARG1 (ptr_t)h, (ptr_t)(h+1));
	    }
	}
	h++;
    }
    if ((ptr_t)h != top) {
	if ((*dirty_fn)(h)) {
            (*push_fn)(SGC_MYENV_ARG1 (ptr_t)h, top);
        }
    }
    if (GC_my_mark_stack.top >=
	GC_my_mark_stack.buf + GC_my_mark_stack.size) {
        ABORT("unexpected mark stack overflow (GC_push_dirty)");
    }
}

# ifndef SMALL_CONFIG
void GC_push_conditional(SGC_MYENV_ARG1 bottom, top, all)
SGC_MYENV_ARG_DECL2
ptr_t bottom;
ptr_t top;
{
    if (all) {
      if (GC_dirty_maintained) {
#	ifdef PROC_VDB
	    /* Pages that were never dirtied cannot contain pointers	*/
	    GC_push_dirty(SGC_MYENV_ARG1 bottom, top, GC_page_was_ever_dirty, GC_push_all);
#	else
	    GC_push_all(SGC_MYENV_ARG1 bottom, top);
#	endif
      } else {
      	GC_push_all(SGC_MYENV_ARG1 bottom, top);
      }
    } else {
	GC_push_dirty(SGC_MYENV_ARG1 bottom, top, GC_page_was_dirty, GC_push_all);
    }
}
#endif

/* Called by App/GC proc. */
/* see xxx_mach_dep.s or mach_dep.c */
# ifdef MSWIN32
  void __cdecl GC_push_one(p)
# else
  void GC_push_one(p)
# endif
word p;
{
/*    DECLARE_MYENV */
    DECLARE_MYENV3
    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
    register ptr_t least_ha = GC_least_plausible_heap_addr;
    GC_PUSH_ONE_STACK(p, least_ha, greatest_ha);
}

# ifdef __STDC__
#   define BASE(p) (word)GC_base((void *)(p))
# else
#   define BASE(p) (word)GC_base((char *)(p))
# endif

/* Called by App/GC proc. */
/* As above, but argument passed preliminary test. */
void GC_push_one_checked(SGC_MYENV_ARG1 p, interior_ptrs)
SGC_MYENV_ARG_DECL2
register word p;
register GC_bool interior_ptrs;
{
    DECLARE_MYENV
    register word r;
    register hdr * hhdr; 
    register int displ;
  
    GET_HDR(p, hhdr);
    if (IS_FORWARDING_ADDR_OR_NIL(hhdr)) {
        if (hhdr != 0 && interior_ptrs) {
          r = BASE(p);
	  hhdr = HDR(r);
	  displ = BYTES_TO_WORDS(HBLKDISPL(r));
	} else {
	  hhdr = 0;
	}
    } else {
        register map_entry_type map_entry;
        
        displ = HBLKDISPL(p);
        map_entry = MAP_ENTRY((hhdr -> hb_map), displ);
        if (map_entry == OBJ_INVALID) {
          if (interior_ptrs) {
            r = BASE(p);
	    displ = BYTES_TO_WORDS(HBLKDISPL(r));
	    if (r == 0) hhdr = 0;
          } else {
            hhdr = 0;
          }
        } else {
          displ = BYTES_TO_WORDS(displ);
          displ -= map_entry;
          r = (word)((word *)(HBLKPTR(p)) + displ);
        }
    }
    /* If hhdr != 0 then r == GC_base(p), only we did it faster. */
    /* displ is the word index within the block.		 */
    if (hhdr == 0) {
    	if (interior_ptrs) {
	    GC_add_to_black_list_stack(p);
	} else {
	    GC_ADD_TO_BLACK_LIST_NORMAL(p);
	}
    } else {
	/* Test-Test&Set */
	word n_marked = 0;
	word n_hasmarked = 0;
	mse * mark_stack_top_reg = GC_my_mark_stack.top;
	mse * mark_stack_limit =
	    &(GC_my_mark_stack.buf[GC_my_mark_stack.size]);
#ifdef COUNT_CROWD
	hdr * ccprevhdr = NULL;
	int ccon = 0;
#endif
	TEST_AND_MARK(r, hhdr, displ, mark_stack_top_reg, mark_stack_limit);
	GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_MARKED, n_marked);
	GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_HASMARKED, n_hasmarked);
	GC_my_mark_stack.top = mark_stack_top_reg;
    }
}

# ifdef TRACE_BUF

# define TRACE_ENTRIES 1000

struct trace_entry {
    char * kind;
    word gc_no;
    word words_allocd;
    word arg1;
    word arg2;
} GC_trace_buf[TRACE_ENTRIES];

int GC_trace_buf_ptr = 0;

void GC_add_trace_entry(char *kind, word arg1, word arg2)
{
    GC_trace_buf[GC_trace_buf_ptr].kind = kind;
    GC_trace_buf[GC_trace_buf_ptr].gc_no = GC_gc_no;
    GC_trace_buf[GC_trace_buf_ptr].words_allocd = GC_words_allocd;
    GC_trace_buf[GC_trace_buf_ptr].arg1 = arg1 ^ 0x80000000;
    GC_trace_buf[GC_trace_buf_ptr].arg2 = arg2 ^ 0x80000000;
    GC_trace_buf_ptr++;
    if (GC_trace_buf_ptr >= TRACE_ENTRIES) GC_trace_buf_ptr = 0;
}

void GC_print_trace(word gc_no, GC_bool lock)
{
    int i;
    struct trace_entry *p;
    
    if (lock) LOCK();
    for (i = GC_trace_buf_ptr-1; i != GC_trace_buf_ptr; i--) {
    	if (i < 0) i = TRACE_ENTRIES-1;
    	p = GC_trace_buf + i;
    	if (p -> gc_no < gc_no || p -> kind == 0) return;
    	printf("Trace:%s (gc:%d,words:%d) 0x%X, 0x%X\n",
    		p -> kind, p -> gc_no, p -> words_allocd,
    		(p -> arg1) ^ 0x80000000, (p -> arg2) ^ 0x80000000);
    }
    printf("Trace incomplete\n");
    if (lock) UNLOCK();
}

# endif /* TRACE_BUF */

#if 0
/*
 * A version of GC_push_all that treats all interior pointers as valid
 */
void GC_push_all_stack(SGC_MYENV_ARG1 bottom, top)
SGC_MYENV_ARG_DECL2
ptr_t bottom;
ptr_t top;
{
# ifdef ALL_INTERIOR_POINTERS
    GC_push_all(SGC_MYENV_ARG1 bottom, top);
#   ifdef TRACE_BUF
        GC_add_trace_entry("GC_push_all_stack", bottom, top);
#   endif
# else
    word * b = (word *)(((long) bottom + ALIGNMENT-1) & ~(ALIGNMENT-1));
    word * t = (word *)(((long) top) & ~(ALIGNMENT-1));
    register word *p;
    register word q;
    register word *lim;
    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
    register ptr_t least_ha = GC_least_plausible_heap_addr;

    if (top == 0) return;
    /* check all pointers in range and put in push if they appear */
    /* to be valid.						  */
      lim = t - 1 /* longword */;
      for (p = b; p <= lim; p = (word *)(((char *)p) + ALIGNMENT)) {
	q = *p;
	GC_PUSH_ONE_STACK(q, least_ha, greatest_ha);
      }
# endif
}
#endif

#ifndef SMALL_CONFIG
/* Push all objects reachable from marked objects in the given block */
/* of size 1 objects.						     */
void GC_push_marked1(SGC_MYENV_ARG1 h, hhdr)
SGC_MYENV_ARG_DECL2
struct hblk *h;
register hdr * hhdr;
{
    word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p;
    word *plim;
    register int i;
    register word q;
    register word mark_word;
    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
    register ptr_t least_ha = GC_least_plausible_heap_addr;
    
    p = (word *)(h->hb_body);
    plim = (word *)(((word)h) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr++;
	    i = 0;
	    while(mark_word != 0) {
	      if (mark_word & 1) {
	          q = p[i];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	      }
	      i++;
	      mark_word >>= 1;
	    }
	    p += WORDSZ;
	}
}


#ifndef UNALIGNED

/* Push all objects reachable from marked objects in the given block */
/* of size 2 objects.						     */
void GC_push_marked2(SGC_MYENV_ARG1 h, hhdr)
SGC_MYENV_ARG_DECL2
struct hblk *h;
register hdr * hhdr;
{
    word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p;
    word *plim;
    register int i;
    register word q;
    register word mark_word;
    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
    register ptr_t least_ha = GC_least_plausible_heap_addr;
    
    p = (word *)(h->hb_body);
    plim = (word *)(((word)h) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr++;
	    i = 0;
	    while(mark_word != 0) {
	      if (mark_word & 1) {
	          q = p[i];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	          q = p[i+1];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	      }
	      i += 2;
	      mark_word >>= 2;
	    }
	    p += WORDSZ;
	}
}

/* Push all objects reachable from marked objects in the given block */
/* of size 4 objects.						     */
/* There is a risk of mark stack overflow here.  But we handle that. */
/* And only unmarked objects get pushed, so it's not very likely.    */
void GC_push_marked4(SGC_MYENV_ARG1 h, hhdr)
SGC_MYENV_ARG_DECL2
struct hblk *h;
register hdr * hhdr;
{
    word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p;
    word *plim;
    register int i;
    register word q;
    register word mark_word;
    register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
    register ptr_t least_ha = GC_least_plausible_heap_addr;
/* #   define GC_greatest_plausible_heap_addr greatest_ha */
/* #   define GC_least_plausible_heap_addr least_ha */
    
    p = (word *)(h->hb_body);
    plim = (word *)(((word)h) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr++;
	    i = 0;
	    while(mark_word != 0) {
	      if (mark_word & 1) {
	          q = p[i];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	          q = p[i+1];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	          q = p[i+2];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	          q = p[i+3];
	          GC_PUSH_ONE_HEAP(q, least_ha, greatest_ha);
	      }
	      i += 4;
	      mark_word >>= 4;
	    }
	    p += WORDSZ;
	}
/* #   undef GC_greatest_plausible_heap_addr */
/* #   undef GC_least_plausible_heap_addr */
}

#endif /* UNALIGNED */

#endif /* SMALL_CONFIG */

/* Push all objects reachable from marked objects in the given block */
void GC_push_marked(SGC_MYENV_ARG1 h, hhdr)
SGC_MYENV_ARG_DECL2
struct hblk *h;
register hdr * hhdr;
{
    DECLARE_MYENV
    register int sz = hhdr -> hb_sz;
    register word * p;
    register int word_no;
    register word * lim;
    register mse * mark_stack_top_reg;
    register mse * mark_stack_limit = &(GC_my_mark_stack.buf[GC_my_mark_stack.size]);
    
    /* Some quick shortcuts: */
        if (hhdr -> hb_obj_kind == PTRFREE) return;
        if (GC_block_empty(hhdr)/* nothing marked */) return;
#   ifdef GATHERSTATS
        GC_n_rescuing_pages++;
#   endif
    GC_my_objects_are_marked = TRUE;
    if (sz > MAXOBJSZ) {
        lim = (word *)(h + 1);
    } else {
        lim = (word *)(h + 1) - sz;
    }
    
    switch(sz) {
#   if !defined(SMALL_CONFIG)    
     case 1:
       GC_push_marked1(SGC_MYENV_ARG1 h, hhdr);
       break;
#   endif
#   if !defined(SMALL_CONFIG) && !defined(UNALIGNED)
     case 2:
       GC_push_marked2(SGC_MYENV_ARG1 h, hhdr);
       break;
     case 4:
       GC_push_marked4(SGC_MYENV_ARG1 h, hhdr);
       break;
#   endif       
     default:
      mark_stack_top_reg = GC_my_mark_stack.top;
      for (p = (word *)h + HDR_WORDS, word_no = HDR_WORDS; p <= lim;
         p += sz, word_no += sz) {
         /* This ignores user specified mark procs.  This currently	*/
         /* doesn't matter, since marking from the whole object		*/
         /* is always sufficient, and we will eventually use the user	*/
         /* mark proc to avoid any bogus pointers.			*/
         if (mark_bit_from_hdr(hhdr, word_no)) {
           /* Mark from fields inside the object */
             PUSH_OBJ((word *)p, hhdr, mark_stack_top_reg, mark_stack_limit);
#	     ifdef GATHERSTATS
		/* Subtract this object from total, since it was	*/
		/* added in twice.					*/
		GC_myg_gc_stat.composite_in_use -= sz;
#	     endif
	    GC_MARKLOG_PUT(GC_LOGKIND_N_MARKED, 1);
         }
      }
      GC_my_mark_stack.top = mark_stack_top_reg;
    }
}

#ifndef SMALL_CONFIG
/* Test whether any page in the given block is dirty	*/
GC_bool GC_block_was_dirty(h, hhdr)
struct hblk *h;
register hdr * hhdr;
{
    register int sz = hhdr -> hb_sz;
    
    if (sz < MAXOBJSZ) {
         return(GC_page_was_dirty(h));
    } else {
    	 register ptr_t p = (ptr_t)h;
         sz += HDR_WORDS;
         sz = WORDS_TO_BYTES(sz);
         while (p < (ptr_t)h + sz) {
             if (GC_page_was_dirty((struct hblk *)p)) return(TRUE);
             p += HBLKSIZE;
         }
         return(FALSE);
    }
}
#endif /* SMALL_CONFIG */

/* Similar to GC_push_next_marked, but return address of next block	*/
struct hblk * GC_push_next_marked(SGC_MYENV_ARG1 h)
SGC_MYENV_ARG_DECL2
struct hblk *h;
{
    register hdr * hhdr;
    
    h = GC_next_block(h);
    if (h == 0) return(0);
    hhdr = HDR(h);
    GC_push_marked(SGC_MYENV_ARG1 h, hhdr);
    return(h + OBJ_SZ_TO_BLOCKS(hhdr -> hb_sz));
}

#ifndef SMALL_CONFIG
/* Identical to above, but mark only from dirty pages	*/
struct hblk * GC_push_next_marked_dirty(SGC_MYENV_ARG1 h)
SGC_MYENV_ARG_DECL2
struct hblk *h;
{
    register hdr * hhdr = HDR(h);
    
    if (!GC_dirty_maintained) { ABORT("dirty bits not set up"); }
    for (;;) {
        h = GC_next_block(h);
        if (h == 0) return(0);
        hhdr = HDR(h);
#	ifdef STUBBORN_ALLOC
          if (hhdr -> hb_obj_kind == STUBBORN) {
            if (GC_page_was_changed(h) && GC_block_was_dirty(h, hhdr)) {
                break;
            }
          } else {
            if (GC_block_was_dirty(h, hhdr)) break;
          }
#	else
	  if (GC_block_was_dirty(h, hhdr)) break;
#	endif
        h += OBJ_SZ_TO_BLOCKS(hhdr -> hb_sz);
    }
    GC_push_marked(SGC_MYENV_ARG1 h, hhdr);
    return(h + OBJ_SZ_TO_BLOCKS(hhdr -> hb_sz));
}
#endif

/* Similar to above, but for uncollectable pages.  Needed since we	*/
/* do not clear marks for such pages, even for full collections.	*/
struct hblk * GC_push_next_marked_uncollectable(SGC_MYENV_ARG1 h)
SGC_MYENV_ARG_DECL2
struct hblk *h;
{
    register hdr * hhdr = HDR(h);

    for (;;) {
        h = GC_next_block(h);
        if (h == 0) {
	    return(0);
	}
        hhdr = HDR(h);
	if (hhdr -> hb_obj_kind == UNCOLLECTABLE) break;
        h += OBJ_SZ_TO_BLOCKS(hhdr -> hb_sz);
    }
    GC_push_marked(SGC_MYENV_ARG1 h, hhdr);
    return(h + OBJ_SZ_TO_BLOCKS(hhdr -> hb_sz));
}


