/* 
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
/* Endo, June 3, 1999 */

# include <stdio.h>
# include "gc_priv.h"

static int rt_hash GC_PROTO((char *addr)); 
static void add_roots_to_index GC_PROTO((struct roots *p)); 

# if !defined(NO_DEBUGGING)
/* For debugging:	*/
void GC_print_static_roots(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    register int i;
    size_t total = 0;
    
    for (i = 0; i < GC_mya_n_root_sets; i++) {
        GC_printf2("Static roots: From 0x%lx to 0x%lx ",
        	   (unsigned long) GC_mya_static_roots[i].r_start,
        	   (unsigned long) GC_mya_static_roots[i].r_end);
        if (GC_mya_static_roots[i].r_tmp) {
            GC_printf0(" (temporary)\n");
        } else {
            GC_printf0("\n");
        }
        total += GC_mya_static_roots[i].r_end - GC_mya_static_roots[i].r_start;
    }
    GC_printf1("Static roots: Total size: %ld\n", (unsigned long) total);
    if (GC_root_size != total) {
     	GC_printf1("GC_root_size incorrect: %ld!!\n",
     		   (unsigned long) GC_root_size);
    }
}
# endif /* NO_DEBUGGING */

/* Primarily for debugging support:	*/
/* Is the address p in one of the registered static			*/
/* root sections?							*/
GC_bool GC_is_static_root(SGC_MYENV_ARG1 p)
SGC_MYENV_ARG_DECL2
ptr_t p;
{
    DECLARE_MYENV
    static int last_root_set = 0;
    register int i;
    
    
    if (p >= GC_mya_static_roots[last_root_set].r_start
        && p < GC_mya_static_roots[last_root_set].r_end) return(TRUE);
    for (i = 0; i < GC_mya_n_root_sets; i++) {
    	if (p >= GC_mya_static_roots[i].r_start
            && p < GC_mya_static_roots[i].r_end) {
            last_root_set = i;
            return(TRUE);
        }
    }
    return(FALSE);
}

#ifndef MSWIN32
#   define LOG_RT_SIZE 6
#   define RT_SIZE (1 << LOG_RT_SIZE) /* Power of 2, may be != MAX_ROOT_SETS */

    static struct roots * root_index[RT_SIZE];
	/* Hash table header.  Used only to check whether a range is 	*/
	/* already present.						*/

static int rt_hash(addr)
char * addr;
{
    word result = (word) addr;
#   if CPP_WORDSZ > 8*LOG_RT_SIZE
	result ^= result >> 8*LOG_RT_SIZE;
#   endif
#   if CPP_WORDSZ > 4*LOG_RT_SIZE
	result ^= result >> 4*LOG_RT_SIZE;
#   endif
    result ^= result >> 2*LOG_RT_SIZE;
    result ^= result >> LOG_RT_SIZE;
    result &= (RT_SIZE-1);
    return(result);
}

/* Is a range starting at b already in the table? If so return a	*/
/* pointer to it, else NIL.						*/
struct roots * GC_roots_present(b)
char *b;
{
    register int h = rt_hash(b);
    register struct roots *p = root_index[h];
    
    while (p != 0) {
        if (p -> r_start == (ptr_t)b) return(p);
        p = p -> r_next;
    }
    return(FALSE);
}

/* Add the given root structure to the index. */
static void add_roots_to_index(p)
struct roots *p;
{
    register int h = rt_hash(p -> r_start);
    
    p -> r_next = root_index[h];
    root_index[h] = p;
}

# else /* MSWIN32 */

#   define add_roots_to_index(p)

# endif

void GC_add_roots(b, e)
char * b; char * e;
{
    DECLARE_MYENV3
    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_g_alloc_ml, SGC_polling_inner);
    GC_add_roots_inner(SGC_MYENV_ARG1 b, e, FALSE);
    PSM_mutex_unlock(&GC_g_alloc_ml);
}

/* Add [b,e) to the root set.  Adding the same interval a second time	*/
/* is a moderately fast noop, and hence benign.  We do not handle	*/
/* different but overlapping intervals efficiently.  (We do handle	*/
/* them correctly.)							*/
/* Tmp specifies that the interval may be deleted before 		*/
/* reregistering dynamic libraries.					*/ 
void GC_add_roots_inner(SGC_MYENV_ARG1 b, e, tmp)
SGC_MYENV_ARG_DECL2
char * b; char * e;
GC_bool tmp;
{
    DECLARE_MYENV
    struct roots * old;
    /* We exclude GC data structures from root sets.  It's usually safe	*/
    /* to mark from those, but it is a waste of time.			*/
    if ( (ptr_t)b < endGC_arrays && (ptr_t)e > beginGC_arrays) {
        if ((ptr_t)e <= endGC_arrays) {
            if ((ptr_t)b >= beginGC_arrays) return;
            e = (char *)beginGC_arrays;
        } else if ((ptr_t)b >= beginGC_arrays) {
            b = (char *)endGC_arrays;
        } else {
            GC_add_roots_inner(SGC_MYENV_ARG1 b, (char *)beginGC_arrays, tmp);
            GC_add_roots_inner(SGC_MYENV_ARG1 (char *)endGC_arrays, e, tmp);
            return;
        }
    }
#   ifdef MSWIN32
      /* Spend the time to ensure that there are no overlapping	*/
      /* or adjacent intervals.					*/
      /* This could be done faster with e.g. a			*/
      /* balanced tree.  But the execution time here is		*/
      /* virtually guaranteed to be dominated by the time it	*/
      /* takes to scan the roots.				*/
      {
        register int i;
        
        for (i = 0; i < GC_mya_n_root_sets; i++) {
            old = GC_mya_static_roots + i;
            if ((ptr_t)b <= old -> r_end && (ptr_t)e >= old -> r_start) {
                if ((ptr_t)b < old -> r_start) {
                    old -> r_start = (ptr_t)b;
                    GC_root_size += (old -> r_start - (ptr_t)b);
                }
                if ((ptr_t)e > old -> r_end) {
                    old -> r_end = (ptr_t)e;
                    GC_root_size += ((ptr_t)e - old -> r_end);
                }
                old -> r_tmp &= tmp;
                break;
            }
        }
        if (i < GC_mya_n_root_sets) {
          /* merge other overlapping intervals */
            struct roots *other;
            
            for (i++; i < GC_mya_n_root_sets; i++) {
              other = GC_mya_static_roots + i;
              b = (char *)(other -> r_start);
              e = (char *)(other -> r_end);
              if ((ptr_t)b <= old -> r_end && (ptr_t)e >= old -> r_start) {
                if ((ptr_t)b < old -> r_start) {
                    old -> r_start = (ptr_t)b;
                    GC_root_size += (old -> r_start - (ptr_t)b);
                }
                if ((ptr_t)e > old -> r_end) {
                    old -> r_end = (ptr_t)e;
                    GC_root_size += ((ptr_t)e - old -> r_end);
                }
                old -> r_tmp &= other -> r_tmp;
                /* Delete this entry. */
                  GC_root_size -= (other -> r_end - other -> r_start);
                  other -> r_start = static_roots[n_root_sets-1].r_start;
                  other -> r_end = static_roots[n_root_sets-1].r_end;
                                  n_root_sets--;
              }
            }
          return;
        }
      }
#   else
      old = GC_roots_present(b);
      if (old != 0) {
        if ((ptr_t)e <= old -> r_end) /* already there */ return;
        /* else extend */
        GC_root_size += (ptr_t)e - old -> r_end;
        old -> r_end = (ptr_t)e;
        return;
      }
#   endif
    if (GC_mya_n_root_sets == MAX_ROOT_SETS) {
        ABORT("Too many root sets\n");
    }
    GC_mya_static_roots[GC_mya_n_root_sets].r_start = (ptr_t)b;
    GC_mya_static_roots[GC_mya_n_root_sets].r_end = (ptr_t)e;
    GC_mya_static_roots[GC_mya_n_root_sets].r_tmp = tmp;
#   ifndef MSWIN32
      GC_mya_static_roots[GC_mya_n_root_sets].r_next = 0;
#   endif
    add_roots_to_index(GC_mya_static_roots + GC_mya_n_root_sets);
    GC_root_size += (ptr_t)e - (ptr_t)b;
    GC_mya_n_root_sets++;
}

void GC_clear_roots GC_PROTO((void))
{
    DECLARE_MYENV
    DECLARE_MYENV3
    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_g_alloc_ml, SGC_polling_inner);
    GC_mya_n_root_sets = 0;
    GC_root_size = 0;
#   ifndef MSWIN32
    {
    	register int i;
    	
    	for (i = 0; i < RT_SIZE; i++) root_index[i] = 0;
    }
#   endif
    PSM_mutex_unlock(&GC_g_alloc_ml);
}

/* Internal use only; lock held.	*/
void GC_remove_tmp_roots(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    register int i;
    
    for (i = 0; i < GC_mya_n_root_sets; ) {
    	if (GC_mya_static_roots[i].r_tmp) {
    	    GC_root_size -= (GC_mya_static_roots[i].r_end -
			     GC_mya_static_roots[i].r_start);
    	    GC_mya_static_roots[i].r_start =
		GC_mya_static_roots[GC_mya_n_root_sets-1].r_start;
    	    GC_mya_static_roots[i].r_end = 
		GC_mya_static_roots[GC_mya_n_root_sets-1].r_end;
    	    GC_mya_static_roots[i].r_tmp = 
		GC_mya_static_roots[GC_mya_n_root_sets-1].r_tmp;
    	    GC_mya_n_root_sets--;
    	} else {
    	    i++;
    	}
    }
#   ifndef MSWIN32
    {
/*    	register int i; */
    	
    	for (i = 0; i < RT_SIZE; i++) root_index[i] = 0;
    	for (i = 0; i < GC_mya_n_root_sets; i++)
	    add_roots_to_index(GC_mya_static_roots + i);
    }
#   endif
    
}

ptr_t GC_approx_sp()
{
    word dummy;
    
    return((ptr_t)(&dummy));
}

/*****************************/
/* called by App/SPMD processes. */
void GC_emit_roots(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    register int i;

    /*
     * push registers - i.e., call GC_push_one(r) for each
     * register contents r.
     */
        GC_push_regs(); /* usually defined in machine_dep.c */
        
    /*
     * Next push static data.  This must happen early on, since it's
     * not robust against mark stack overflow.
     */
     /* Reregister dynamic libraries, in case one got added.	*/
#      if (defined(DYNAMIC_LOADING) || defined(MSWIN32) || defined(PCR)) \
           && !defined(SRC_M3)
         GC_remove_tmp_roots(SGC_MYENV_ARG0);
        GC_register_dynamic_libraries(SGC_MYENV_ARG0);
#      endif

     /* Mark everything in static data areas                             */
       for (i = 0; i < GC_mya_n_root_sets; i++) {
	   GC_push_all_contents(SGC_MYENV_ARG1 GC_mya_static_roots[i].r_start,
		       GC_mya_static_roots[i].r_end);
       }

    /*
     * Now traverse stacks.
     */
    GC_push_all_stack_contents(SGC_MYENV_ARG0);
}

/* called by GC/SPMD processes. */
/* GC processes steal mark stack entries from App procs. */
void GC_push_roots(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int n;
    int i;
    ptr_t *p;

    if (GC_np_fixed == 0) {
	/* GC process in NP-variable model */
	if (GC_my_envp == GC_g_gcleader_envp) {
	    GC_thread envp;
	    mse *pfrom;
	    mse *pto;
	    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
		if (GC_l_initialized(envp) == 0) continue;
		if (GC_l_terminated(envp) != 0) continue;
		/* steal mark stack entries from App procs */
		pfrom = GC_l_mark_stack(envp).bottom;
		pto = GC_my_mark_stack.top+1;
		n = GC_l_mark_stack(envp).top
		    - GC_l_mark_stack(envp).bottom + 1;
		for (i = 0; i < n; i++) {
		    COPY_MSE(pto, pfrom);
		    pto++;
		    pfrom++;
		}
		GC_my_mark_stack.top = pto-1;
		GC_l_mark_stack(envp).top = GC_l_mark_stack(envp).bottom - 1;
	    }
	}
    }
    else {
	/* SPMD process in NP-fixed model */
	/* do nothing */
    }

    return;
}
