/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, February 8, 1999 */

#include "gc_priv.h"

#ifdef SUNOS5

#include <unistd.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/procfs.h>
#include <sys/lwp.h>

int GC_sh_processor_bind(int procno)
{
    int n_processor;
    int pr;
    int rc;

    int vpr;
    int n;
    
    n_processor = sysconf(_SC_NPROCESSORS_ONLN);
    vpr = procno % n_processor;
    /* Lookup vpr_th alive processor in non-elegant method... */
    n = 0;
    for (pr = 0; pr < 0x1000; pr++) {
	processor_info_t info;
	rc = processor_info(pr, &info);
	if (rc == 0 && info.pi_state == P_ONLINE) {
	    if (n == vpr) {
		break;
	    }
	    n++;
	}
    }
    
    rc = processor_bind(P_LWPID, P_MYID, pr, NULL);

    if (rc == 0) {
	return pr;
    }
    else {
	return -1;
    }
}

int GC_sh_processor_bind_currentpe()
{
    char buf[32];
    int fd;
    int fd2;
    int pr = 0;
    int rc;
    /* get current processor id */
    sprintf(buf, "/proc/%d", (int)getpid());
    if ((fd = open (buf, O_RDONLY)) >= 0) {
	lwpid_t me = _lwp_self();
	if ((fd2 = ioctl(fd, PIOCOPENLWP, &me)) >= 0) {
	    struct prstatus prstatus;
	    if (ioctl (fd2, PIOCSTATUS, &prstatus) >= 0) {
		pr = prstatus.pr_processor;
	    }
	    close(fd2);
	}
	else {
	    GC_err_printf0("ioctl failed\n");
	    close(fd);
	    return -1;
	}
	close(fd);
    }
    else {
	GC_err_printf1("open failed %s\n", buf);
	return -1;
    }
    /* bind to a processer */
    rc = processor_bind(P_PID, P_MYID, pr, NULL);
GC_err_printf2("bound to %d, rc=%d\n", pr, rc);
    if (rc == 0) {
	return pr;
    }
    else {
	return -1;
    }
}

int GC_sh_processor_unbind()
{
    processor_bind(P_PID, P_MYID, PBIND_NONE, NULL);
    return 0;
}

#elif defined IRIX5

#include <unistd.h>
#include <sys/sysmp.h>
#include <sys/sysinfo.h>

int GC_sh_processor_bind(int procno)
{
    int n_processor;
    int pr;
    int rc;
    n_processor = sysmp(MP_NPROCS);
    if (n_processor <= 0) {
	GC_err_printf0("!!Failed to sysmp(MP_NPROCS)\n");
    }

    pr = procno % n_processor;
    rc = sysmp(MP_MUSTRUN, pr);

    if (rc == 0) {
	return pr;
    }
    else {
	return -1;
    }
}

int GC_sh_processor_bind_currentpe()
{
    return -1;
}

int GC_sh_processor_unbind()
{
    sysmp(MP_RUNANYWHERE);
    return 0;
}

#endif

/* save call chain for Intel (by kamo) */

#if defined(I386) && defined(SUNOS5)

#   include <sys/procfs.h>
#   include <sys/frame.h>
#   if NARGS > 6
--> We only know how to to get the first 6 arguments
#   endif

#ifdef SAVE_CALL_CHAIN
/* Fill in the pc and argument information for up to NFRAMES of my	*/
/* callers.  Ignore my frame and my callers frame.			*/

word GC_save_regs_in_stack(){
  word dummy_a;
  struct frame *f_a;
  struct frame *g_a;

  f_a = (struct frame*)(&dummy_a + 1);
  g_a = (struct frame*)(f_a -> fr_savfp);
  g_a = (struct frame*)(f_a -> fr_savfp);

  return (word)(f_a);
}

void GC_save_callers (info) 
struct callinfo info[NFRAMES];
{
  struct frame *frame;
  struct frame *fp;
  int nframes = 0;
/*  word GC_save_regs_in_stack();*/
  word *args;

  frame = (struct frame *) GC_save_regs_in_stack ();
  
  for (fp = (struct frame *)frame -> fr_savfp; 
       fp != 0 && nframes < NFRAMES;
       fp = (struct frame *)fp -> fr_savfp, nframes++) {
      register int i;
      
      info[nframes].ci_pc = fp->fr_savpc;
      /*      GC_err_printf2("FP = %016lx, SP = %016lx\n", fp->fr_savfp, fp->fr_savpc); */
      args = (word *)(fp + 1);
      for (i = 0; i < NARGS; i++) {
	info[nframes].ci_arg[i] = ~(args[i]);
      }
  }
  if (nframes < NFRAMES) info[nframes].ci_pc = 0;
}

#endif /* SAVE_CALL_CHAIN */
#endif /* 386 */

#ifdef SAVE_CALL_CHAIN

void GC_print_callers (info)
struct callinfo info[NFRAMES];
{
    register int i,j;
    
    GC_err_printf0("\tCall chain at allocation:\n");
    for (i = 0; i < NFRAMES; i++) {
     	if (info[i].ci_pc == 0) break;
     	GC_err_printf0("\t\targs: ");
     	for (j = 0; j < NARGS; j++) {
     	    if (j != 0) GC_err_printf0(", ");
     	    GC_err_printf2("%d (0x%X)", ~(info[i].ci_arg[j]),
     	    				~(info[i].ci_arg[j]));
     	}
     	GC_err_printf1("\n\t\t##PC##= 0x%X\n", info[i].ci_pc);
    }
}

#endif /* SAVE_CALL_CHAIN */


