/* 
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
/* Endo, June 3, 1999 */

#include <stdio.h>
#include "gc_priv.h"

static int set_bits GC_PROTO((word n)); 

# ifdef FIND_LEAK
static void report_leak GC_PROTO((struct hblk *, int));

# if defined(__STDC__) || defined(__cplusplus)
static void report_leak(struct hblk *hblk, int word_no)
#else
static void report_leak(hblk, word_no)
  struct hblk *hblk;
  int word_no;
#endif
{
    DECLARE_MYENV
    DECLARE_MYENV3
    ptr_t p;
    word sz;
    int ok;
    hdr *hhdr;

    p = (ptr_t)((long)hblk + WORDS_TO_BYTES(word_no));
    hhdr = HDR(p);
    ok = hhdr->hb_obj_kind;
    sz = hhdr->hb_sz;
    if(!GC_myg_pr.freelist_top[ok][sz]){
	GC_myg_pr.freelist_top[ok][sz] = p;
    }
    if (ok == PTRFREE) {
        GC_err_printf0("Leaked atomic object at ");
    } else {
        GC_err_printf0("Leaked composite object at ");
    }
    if (GC_debugging_started && GC_has_debug_info(p)) {
        GC_print_obj(p);
    } else {
        GC_err_printf2("0x%lx (appr. size = %ld)\n",
       		      (unsigned long)p,
       		      (unsigned long)WORDS_TO_BYTES(sz));
    }
}

#   define FOUND_FREE(hblk, word_no) \
      if (abort_if_found) { \
         report_leak(hblk, word_no); \
      }
# else
#   define FOUND_FREE(hblk, word_no)
# endif

/*
 * reclaim phase
 *
 */


/*
 * Test whether a block is completely empty, i.e. contains no marked
 * objects.  This does not require the block to be in physical
 * memory.
 */
 
# if defined(__STDC__) || defined(__cplusplus)
GC_bool GC_block_empty(register hdr *hhdr)
#else
GC_bool GC_block_empty(hhdr)
register hdr * hhdr;
#endif
{
    register word *p = (word *)(&(hhdr -> hb_marks[0]));
    register word * plim =
	    		(word *)(&(hhdr -> hb_marks[MARK_BITS_SZ]));
    while (p < plim) {
	if (*p) return(FALSE);
	p += 1;
    }
    return(TRUE);
}

# ifdef GATHERSTATS
#   define INCR_WORDS(sz) n_words_found += (sz)
# else
#   define INCR_WORDS(sz)
# endif
/*
 * Restore unmarked small objects in h of size sz to the object
 * free list.  Returns the new list.
 * Clears unmarked objects.
 */
/*ARGSUSED*/
# if defined(__STDC__) || defined(__cplusplus)
ptr_t GC_reclaim_clear(
		       register struct hblk *hbp,	/* ptr to current heap block		*/
		       register hdr * hhdr,
		       register word sz,
		       register ptr_t list,
		       GC_bool abort_if_found)		/* Abort if a reclaimable object is found */
#else
ptr_t GC_reclaim_clear(hbp, hhdr, sz, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
register hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
register word sz;
#endif
{
    register int word_no;
    register word *p, *q, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif        
    
    p = (word *)(hbp->hb_body);
    word_no = HDR_WORDS;
    plim = (word *)((((word)hbp) + HBLKSIZE)
		   - WORDS_TO_BYTES(sz));

    /* go through all words in block */
	while( p <= plim )  {
	    if( mark_bit_from_hdr(hhdr, word_no) ) {
		p += sz;
	    } else {
		FOUND_FREE(hbp, word_no);
		INCR_WORDS(sz);
		/* object is available - put on list */
		    obj_link(p) = list;
		    list = ((ptr_t)p);
		/* Clear object, advance p to next object in the process */
		    q = p + sz;
                    p++; /* Skip link field */
                    while (p < q) {
			*p++ = 0;
		    }
	    }
	    word_no += sz;
	}
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
}

#ifndef SMALL_CONFIG

/*
 * A special case for 2 word composite objects (e.g. cons cells):
 */
/*ARGSUSED*/
ptr_t GC_reclaim_clear2(hbp, hhdr, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
{
    register word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif
    register word mark_word;
    register int i;
#   define DO_OBJ(start_displ) \
	if (!(mark_word & ((word)1 << start_displ))) { \
	    FOUND_FREE(hbp, p - (word *)hbp + start_displ); \
	    p[start_displ] = (word)list; \
	    list = (ptr_t)(p+start_displ); \
	    p[start_displ+1] = 0; \
	    INCR_WORDS(2); \
	}
    
    p = (word *)(hbp->hb_body);
    plim = (word *)(((word)hbp) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr;
	    mark_word_addr += 1;
	    for (i = 0; i < WORDSZ; i += 8) {
		DO_OBJ(0);
		DO_OBJ(2);
		DO_OBJ(4);
		DO_OBJ(6);
		p += 8;
		mark_word >>= 8;
	    }
	}	        
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
#   undef DO_OBJ
}

/*
 * Another special case for 4 word composite objects:
 */
/*ARGSUSED*/
ptr_t GC_reclaim_clear4(hbp, hhdr, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
{
    register word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif
    register word mark_word;
#   define DO_OBJ(start_displ) \
	if (!(mark_word & ((word)1 << start_displ))) { \
	    FOUND_FREE(hbp, p - (word *)hbp + start_displ); \
	    p[start_displ] = (word)list; \
	    list = (ptr_t)(p+start_displ); \
	    p[start_displ+1] = 0; \
	    p[start_displ+2] = 0; \
	    p[start_displ+3] = 0; \
	    INCR_WORDS(4); \
	}
    
    p = (word *)(hbp->hb_body);
    plim = (word *)(((word)hbp) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr;
	    mark_word_addr += 1;
	    DO_OBJ(0);
	    DO_OBJ(4);
	    DO_OBJ(8);
	    DO_OBJ(12);
	    DO_OBJ(16);
	    DO_OBJ(20);
	    DO_OBJ(24);
	    DO_OBJ(28);
#	    if CPP_WORDSZ == 64
	      DO_OBJ(32);
	      DO_OBJ(36);
	      DO_OBJ(40);
	      DO_OBJ(44);
	      DO_OBJ(48);
	      DO_OBJ(52);
	      DO_OBJ(56);
	      DO_OBJ(60);
#	    endif
	    p += WORDSZ;
	}	        
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
#   undef DO_OBJ
}

#endif /* !SMALL_CONFIG */

/* The same thing, but don't clear objects: */
/*ARGSUSED*/
ptr_t GC_reclaim_uninit(hbp, hhdr, sz, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
register hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
register word sz;
{
    register int word_no;
    register word *p, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif
    
    p = (word *)(hbp->hb_body);
    word_no = HDR_WORDS;
    plim = (word *)((((word)hbp) + HBLKSIZE)
		   - WORDS_TO_BYTES(sz));

    /* go through all words in block */
	while( p <= plim )  {
	    if( !mark_bit_from_hdr(hhdr, word_no) ) {
		FOUND_FREE(hbp, word_no);
		INCR_WORDS(sz);
		/* object is available - put on list */
		    obj_link(p) = list;
		    list = ((ptr_t)p);
	    }
	    p += sz;
	    word_no += sz;
	}
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
}

#ifndef SMALL_CONFIG
/*
 * Another special case for 2 word atomic objects:
 */
/*ARGSUSED*/
ptr_t GC_reclaim_uninit2(hbp, hhdr, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
{
    register word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif
    register word mark_word;
    register int i;
#   define DO_OBJ(start_displ) \
	if (!(mark_word & ((word)1 << start_displ))) { \
	    FOUND_FREE(hbp, p - (word *)hbp + start_displ); \
	    p[start_displ] = (word)list; \
	    list = (ptr_t)(p+start_displ); \
	    INCR_WORDS(2); \
	}
    
    p = (word *)(hbp->hb_body);
    plim = (word *)(((word)hbp) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr;
	    mark_word_addr += 1;
	    for (i = 0; i < WORDSZ; i += 8) {
		DO_OBJ(0);
		DO_OBJ(2);
		DO_OBJ(4);
		DO_OBJ(6);
		p += 8;
		mark_word >>= 8;
	    }
	}	        
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
#   undef DO_OBJ
}

/*
 * Another special case for 4 word atomic objects:
 */
/*ARGSUSED*/
ptr_t GC_reclaim_uninit4(hbp, hhdr, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
{
    register word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif
    register word mark_word;
#   define DO_OBJ(start_displ) \
	if (!(mark_word & ((word)1 << start_displ))) { \
	    FOUND_FREE(hbp, p - (word *)hbp + start_displ); \
	    p[start_displ] = (word)list; \
	    list = (ptr_t)(p+start_displ); \
	    INCR_WORDS(4); \
	}
    
    p = (word *)(hbp->hb_body);
    plim = (word *)(((word)hbp) + HBLKSIZE);

    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr;
	    mark_word_addr += 1;
	    DO_OBJ(0);
	    DO_OBJ(4);
	    DO_OBJ(8);
	    DO_OBJ(12);
	    DO_OBJ(16);
	    DO_OBJ(20);
	    DO_OBJ(24);
	    DO_OBJ(28);
#	    if CPP_WORDSZ == 64
	      DO_OBJ(32);
	      DO_OBJ(36);
	      DO_OBJ(40);
	      DO_OBJ(44);
	      DO_OBJ(48);
	      DO_OBJ(52);
	      DO_OBJ(56);
	      DO_OBJ(60);
#	    endif
	    p += WORDSZ;
	}	        
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
#   undef DO_OBJ
}

/* Finally the one word case, which never requires any clearing: */
/*ARGSUSED*/
ptr_t GC_reclaim1(hbp, hhdr, list, abort_if_found)
register struct hblk *hbp;	/* ptr to current heap block		*/
hdr * hhdr;
GC_bool abort_if_found;		/* Abort if a reclaimable object is found */
register ptr_t list;
{
    register word * mark_word_addr = &(hhdr->hb_marks[divWORDSZ(HDR_WORDS)]);
    register word *p, *plim;
#   ifdef GATHERSTATS
        register int n_words_found = 0;
#   endif
    register word mark_word;
    register int i;
#   define DO_OBJ(start_displ) \
	if (!(mark_word & ((word)1 << start_displ))) { \
	    FOUND_FREE(hbp, p - (word *)hbp + start_displ); \
	    p[start_displ] = (word)list; \
	    list = (ptr_t)(p+start_displ); \
	    INCR_WORDS(1); \
	}

    p = (word *)(hbp->hb_body);
    plim = (word *)(((word)hbp) + HBLKSIZE);
    /* go through all words in block */
	while( p < plim )  {
	    mark_word = *mark_word_addr;
	    mark_word_addr += 1;
	    for (i = 0; i < WORDSZ; i += 4) {
		DO_OBJ(0);
		DO_OBJ(1);
		DO_OBJ(2);
		DO_OBJ(3);
		p += 4;
		mark_word >>= 4;
	    }
	}	        
#   ifdef GATHERSTATS
    {
	DECLARE_MYENV
	DECLARE_MYENV3
	GC_myg_gc_stat.mem_found += n_words_found;
    }
#   endif
    return(list);
#   undef DO_OBJ
}

#endif /* !SMALL_CONFIG */

/*
 * Restore unmarked small objects in the block pointed to by hbp
 * to the appropriate object free list.
 * If entirely empty blocks are to be completely deallocated, then
 * caller should perform that check.
 */
void GC_reclaim_small_nonempty_block(SGC_MYENV_ARG1 hbp, abort_if_found)
SGC_MYENV_ARG_DECL2
register struct hblk *hbp;	/* ptr to current heap block		*/
int abort_if_found;		/* Abort if a reclaimable object is found */
{
    DECLARE_MYENV
    hdr * hhdr;
    register word sz;		/* size of objects in current block	*/
    register struct obj_kind * ok;
    register ptr_t * flh;
    register int kind;
    
    hhdr = HDR(hbp);
    sz = hhdr -> hb_sz;
    hhdr -> hb_last_reclaimed = (unsigned short) GC_gc_no;
    kind = hhdr -> hb_obj_kind;
    ok = &GC_obj_kinds[kind];
    flh = &(GC_mya_freelist[kind][sz]);
    GC_write_hint(hbp);

    if (ok -> ok_init) {
      switch(sz) {
#      ifndef SMALL_CONFIG
        case 1:
            *flh = GC_reclaim1(hbp, hhdr, *flh, abort_if_found);
            break;
        case 2:
            *flh = GC_reclaim_clear2(hbp, hhdr, *flh, abort_if_found);
            break;
        case 4:
            *flh = GC_reclaim_clear4(hbp, hhdr, *flh, abort_if_found);
            break;
#      endif
        default:
            *flh = GC_reclaim_clear(hbp, hhdr, sz, *flh, abort_if_found);
            break;
      }
    } else {
      switch(sz) {
#      ifndef SMALL_CONFIG
        case 1:
            *flh = GC_reclaim1(hbp, hhdr, *flh, abort_if_found);
            break;
        case 2:
            *flh = GC_reclaim_uninit2(hbp, hhdr, *flh, abort_if_found);
            break;
        case 4:
            *flh = GC_reclaim_uninit4(hbp, hhdr, *flh, abort_if_found);
            break;
#      endif
        default:
            *flh = GC_reclaim_uninit(hbp, hhdr, sz, *flh, abort_if_found);
            break;
      }
    } 
    if (IS_UNCOLLECTABLE(kind)) GC_set_hdr_marks(hhdr);
}

/*
 * Restore an unmarked large object or an entirely empty blocks of small objects
 * to the heap block free list.
 * Otherwise enqueue the block for later processing
 * by GC_reclaim_small_nonempty_block.
 * If abort_if_found is TRUE, then process any block immediately.
 */
void GC_reclaim_block(SGC_MYENV_ARG_DECL1 
		      struct hblk *hbp /* ptr to current heap block */, 
		      word abort_if_found /* Abort if a reclaimable object is found */)
{
    DECLARE_MYENV
    register hdr * hhdr;
    register word sz;		/* size of objects in current block	*/
    register struct obj_kind * ok;
    GC_thread home_envp;
#ifndef DO_PRECLAIM
    struct hblk ** rlh;
#endif

    hhdr = HDR(hbp);
    sz = hhdr -> hb_sz;
    ok = &GC_obj_kinds[hhdr -> hb_obj_kind];

    if( sz > MAXOBJSZ ) {  /* 1 big object */
        if( !mark_bit_from_hdr(hhdr, HDR_WORDS) ) {
	    FOUND_FREE(hbp, HDR_WORDS);
#	    ifdef GATHERSTATS
	        GC_myg_gc_stat.mem_found += sz;
#	    endif
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_FREED, 1);
#ifdef DO_PRECLAIM
	    GC_freehblk_local(SGC_MYENV_ARG1 hbp);
#else
	    GC_freehblk(hbp);
#endif
	}
    } else {
        GC_bool empty = GC_block_empty(hhdr);
        if (abort_if_found) {
	    GC_reclaim_small_nonempty_block(SGC_MYENV_ARG1 hbp, 
					    (int)abort_if_found);
#ifdef FIND_LEAK
	    GC_flush_temp_flist(SGC_MYENV_ARG1 hhdr);
#endif /* FIND_LEAK */
        } else if (empty) {
#	  ifdef GATHERSTATS
	    GC_myg_gc_stat.mem_found += BYTES_TO_WORDS(HBLKSIZE);
#	  endif
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_FREED, 1);
#ifdef DO_PRECLAIM
	    GC_freehblk_local(SGC_MYENV_ARG1 hbp);
#else
	    GC_freehblk(hbp);
#endif
        } else {
	    /* group of smaller objects, enqueue the real work */
	    home_envp = hhdr->home_envp;
#ifdef DO_PRECLAIM
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_TORECLAIM, 1);
	    /* for Parallel reclaim */

	    if (home_envp != GC_myg_pr.temp_rlist_home_envp) {
		GC_flush_temp_rlist(SGC_MYENV_ARG0);
		GC_myg_pr.temp_rlist_home_envp = home_envp;
	    }
	    
	    hhdr->hb_next = GC_myg_pr.temp_rlist_start;
	    GC_myg_pr.temp_rlist_start = hbp;
#else /* DO_PRECLAIM */
	    GC_FAST_MARKLOG_PUT(GC_LOGKIND_N_TORECLAIM, 1);
	    rlh = &(GC_la_reclaim_list(home_envp)[hhdr -> hb_obj_kind][sz]);
	    hhdr -> hb_next = *rlh;
	    *rlh = hbp;
#endif /* DO_PRECLAIM */
        }
    }
}

#if !defined(NO_DEBUGGING)
/* Routines to gather and print heap block info 	*/
/* intended for debugging.  Otherwise should be called	*/
/* with lock.						*/
static int number_of_blocks;
static int total_bytes;

/* Number of set bits in a word.  Not performance critical.	*/
static int set_bits(n)
word n;
{
    register word m = n;
    register int result = 0;
    
    while (m > 0) {
    	if (m & 1) result++;
    	m >>= 1;
    }
    return(result);
}

/* Return the number of set mark bits in the given header	*/
int GC_n_set_marks(hhdr)
hdr * hhdr;
{
    register int result = 0;
    register int i;
    
    for (i = 0; i < MARK_BITS_SZ; i++) {
        result += set_bits(hhdr -> hb_marks[i]);
    }
    return(result);
}

/*ARGSUSED*/
void GC_print_block_descr(h, dummy)
struct hblk *h;
word dummy;
{
    register hdr * hhdr = HDR(h);
    register int bytes = WORDS_TO_BYTES(hhdr -> hb_sz);
    
    GC_printf3("(%lu:%lu,%lu)", (unsigned long)(hhdr -> hb_obj_kind),
    			        (unsigned long)bytes,
    			        (unsigned long)(GC_n_set_marks(hhdr)));
    bytes += HDR_BYTES + HBLKSIZE-1;
    bytes &= ~(HBLKSIZE-1);
    total_bytes += bytes;
    number_of_blocks++;
}

void GC_print_block_list()
{
    GC_printf0("(kind(0=ptrfree,1=normal,2=unc.,3=stubborn):size_in_bytes, #_marks_set)\n");
    number_of_blocks = 0;
    total_bytes = 0;
    GC_apply_to_all_blocks(GC_print_block_descr, (word)0);
    GC_printf2("\nblocks = %lu, bytes = %lu\n",
    	       (unsigned long)number_of_blocks,
    	       (unsigned long)total_bytes);
}

#endif /* NO_DEBUGGING */

ptr_t GC_sh_after_marksweep_local(ptr_t arg1, ptr_t arg2, ptr_t arg3);

/* Called by App proc. */
/*
 * Sweep blocks of the indicated object size and kind until either the
 * appropriate free list is nonempty, or there are no more blocks to
 * sweep.
 */
void GC_continue_reclaim(SGC_MYENV_ARG1 sz, kind)
SGC_MYENV_ARG_DECL2
word sz;	/* words */
int kind;
{
    DECLARE_MYENV
    register hdr * hhdr;
    register struct hblk * hbp;
    struct hblk ** rlh = GC_mya_reclaim_list[kind];
    ptr_t *flh = &(GC_mya_freelist[kind][sz]);
    
    if (rlh == 0) return;	/* No blocks of this kind.	*/
    rlh += sz;
    while ((hbp = *rlh) != 0) {
        hhdr = HDR(hbp);
        *rlh = hhdr -> hb_next;
        GC_reclaim_small_nonempty_block(SGC_MYENV_ARG1 hbp, FALSE);
        if (*flh != 0) break;
    }
}

/* Called by App proc. */
/* Analogous to GC_continue_reclaim, but this function steals */
/* a block from global-reclaim-list. */
void GC_continue_reclaim_g(SGC_MYENV_ARG1 sz, kind)
SGC_MYENV_ARG_DECL2
word sz;	/* words */
int kind;
{
    DECLARE_MYENV
    register hdr * hhdr;
    register struct hblk * hbp;
    struct hblk ** rlh;
    ptr_t *flh = &(GC_mya_freelist[kind][sz]);
    
#ifdef FIND_LEAK
    ptr_t gfl;
    /* global free list */
    if(GC_g_freelist[kind] != 0){
	while(1){
	    gfl = GC_g_freelist[kind][sz];
	    if(gfl == 0) break;
	    hhdr = HDR(gfl);
	    if(hhdr == 0){
		ABORT("global freelist points invalid block\n");
	    }
	    PSM_mutex_lock(&GC_g_freelist_ml);
	    PSM_mutex_lock(&hhdr->gfl_ml);
	    if((gfl == GC_g_freelist[kind][sz]) &&
	       (gfl == hhdr->gfl)){
		/* success */
		/*PSM_mutex_lock(&GC_mya_alloc_ml);*/
		/*
		GC_err_printf6("gfl_top:%08x, "
			       "*gfl_top:%08x, "
			       "gfl:%08x, "
			       "*gfl:%08x, "
			       "flh:%08x, "
			       "*flh:%08x\n",
			       hhdr->gfl_top, *(ptr_t *)(hhdr->gfl_top),
			       hhdr->gfl, *(ptr_t *)(hhdr->gfl),
			       flh, *(ptr_t *)(flh));
		*/
		GC_g_freelist[kind][sz] = *(ptr_t *)(hhdr->gfl_top);
		*(ptr_t *)(hhdr->gfl_top) = *flh;
		*flh = gfl;
		hhdr->gfl = 0;
		hhdr->gfl_top = 0;
		/*PSM_mutex_unlock(&GC_mya_alloc_ml);*/
		PSM_mutex_unlock(&hhdr->gfl_ml);
		PSM_mutex_unlock(&GC_g_freelist_ml);
		return;
	    }
	    PSM_mutex_unlock(&hhdr->gfl_ml);
	    PSM_mutex_unlock(&GC_g_freelist_ml);
	}
    }
#endif /* FIND_LEAK */
    /* global reclaim list */
    rlh = GC_g_reclaim_list[kind];
    if (rlh == 0) return;	/* No blocks of this kind.	*/
    rlh += sz;
    PSM_mutex_lock(&GC_g_reclaim_list_ml);
    while ((hbp = *rlh) != 0) {
        hhdr = HDR(hbp);
	/* Setup home proc of this page. */
	hhdr -> home_envp = GC_my_envp;
        *rlh = hhdr -> hb_next;
        GC_reclaim_small_nonempty_block(SGC_MYENV_ARG1 hbp, FALSE);
        if (*flh != 0) break;
    }
    PSM_mutex_unlock(&GC_g_reclaim_list_ml);
}

#ifdef FIND_LEAK
#define SAME_BLOCK(x,y) !((((unsigned long)x)^((unsigned long)y))&(~(HBLKSIZE-1)))
ptr_t GC_yield_freelist_inner(ptr_t p, ptr_t *gflp){
    ptr_t top = 0;
    ptr_t bottom;
    hdr *hhdr;

    while(p != 0){
	bottom = p;
	top = 0;
	while((p != 0) && SAME_BLOCK(bottom, p)){
	    top = p;
	    p = *(ptr_t *)p;
	}
	if (top == 0) break;
	hhdr = HDR(top);
	PSM_mutex_lock(&hhdr->gfl_ml);
	if(hhdr->gfl != 0){
	    *(ptr_t *)top = *(ptr_t *)(hhdr->gfl_top);
	    *(ptr_t *)(hhdr->gfl_top) = bottom;
	    hhdr->gfl_top = top;
	}else{
	    hhdr->gfl = bottom;
	    hhdr->gfl_top = top;
	    PSM_mutex_lock(&GC_g_freelist_ml);
	    *(ptr_t *)top = *(ptr_t *)gflp;
	    *(ptr_t *)gflp = bottom;
	    PSM_mutex_unlock(&GC_g_freelist_ml);
	}
	PSM_mutex_unlock(&hhdr->gfl_ml);
    }
    if(top == 0){
	ABORT("invalid freelist\n");
    }
    return top;
}

/* called by terminating app proc. */
/* this function removes the local freelist and append to global-freelist. */
/* don't use GC_my_envp in this function. */
void GC_yield_freelist(GC_thread envp)
{
    ptr_t fl;
    word sz;
    int kind;
    
    for (kind = 0; kind < GC_n_kinds; kind++) {
    	if (GC_la_freelist(envp)[kind] == 0) continue;
    	for (sz = 1; sz <= MAXOBJSZ; sz++) {
	    fl = GC_la_freelist(envp)[kind][sz];
	    if(fl == 0) continue;
	    GC_yield_freelist_inner(fl, &(GC_g_freelist[kind][sz]));
	    GC_la_freelist(envp)[kind][sz] = 0;
	}
    }
}
#endif /* FIND_LEAK */

/* called by terminating app proc. */
/* move blocks in local reclaim list into global-reclaim-list. */
/* don't use GC_my_envp in this function. */
void GC_yield_reclaim_list(GC_thread envp)
{
    register hdr * hhdr;
    register struct hblk * hbp;
    struct hblk ** rlh;
    struct hblk ** grlh;
    word sz;
    int kind;
    
    for (kind = 0; kind < GC_n_kinds; kind++) {
    	if (GC_la_reclaim_list(envp)[kind] == 0) continue;
    	for (sz = 1; sz <= MAXOBJSZ; sz++) {
	    rlh = &GC_la_reclaim_list(envp)[kind][sz];
	    grlh = &GC_g_reclaim_list[kind][sz];
	    
	    while ((hbp = *rlh) != 0) {
		hhdr = HDR(hbp);
		/* Setup home proc of this page. */
		hhdr -> home_envp = NULL;
		*rlh = hhdr -> hb_next;
		
		/* Link to global reclaim list */
		hhdr -> hb_next = *grlh;
		*grlh = hbp;
	    }

	    *rlh = 0;
	}
    }
}

void GC_yield_block(struct hblk *hbp, word data)
{
    GC_thread envp = (GC_thread)data;
    hdr * hhdr;
    hhdr = HDR(hbp);
    if (hhdr->home_envp == envp) {
	/* change hbp `homeless'. */
	hhdr->home_envp = NULL;
    }
    return;
}

/* called by terminating App proc. */
/* don't use GC_my_envp in this function. */
void GC_yield_all_blocks(GC_thread envp)
{
    GC_apply_to_all_blocks(GC_yield_block, (word)envp);
    return;
}

/*
 * Reclaim all blocks that have been recently reclaimed.
 * Clear lists of blocks waiting to be reclaimed.
 * Must be done before clearing mark bits with the world running,
 * since otherwise a subsequent reclamation of block would see
 * the wrong mark bits.  (Alternatively, GC_reclaim_all
 * may be used.)
 * SHOULD PROBABLY BE INCREMENTAL
 */
void GC_reclaim_or_delete_all(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    register word sz;
    register int kind;
    register hdr * hhdr;
    register struct hblk * hbp;
    struct hblk ** rlp;
    struct hblk ** rlh;
#   ifdef PRINTTIMES
	CLOCK_TYPE start_time;
	CLOCK_TYPE done_time;
	
	GET_TIME(start_time);
#   endif
    
    for (kind = 0; kind < GC_n_kinds; kind++) {
    	rlp = GC_mya_reclaim_list[kind];
    	if (rlp == 0) continue;
    	for (sz = 1; sz <= MAXOBJSZ; sz++) {
    	    rlh = rlp + sz;
    	    while ((hbp = *rlh) != 0) {
        	hhdr = HDR(hbp);
        	*rlh = hhdr -> hb_next;
        	if (hhdr -> hb_last_reclaimed == GC_gc_no - 1) {
        	    /* It's likely we'll need it this time, too	*/
        	    /* It's been touched recently, so this	*/
        	    /* shouldn't trigger paging.		*/
        	    GC_reclaim_small_nonempty_block(SGC_MYENV_ARG1 hbp, FALSE);
        	}
            }
        }
    }
#   ifdef PRINTTIMES
	GET_TIME(done_time);
	GC_printf1("Disposing of reclaim lists took %lu msecs\n",
	           MS_TIME_DIFF(done_time,start_time));
#   endif
}

