/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, August 17, 1999 */

#include <stdio.h>
#include "gc_priv.h"
#include <sys/resource.h>

#define SHZZZ_MIN(a,b) ((a < b)? a: b)
#define SHZZZ_CALCDIFFTIME(st, et) \
((et.tv_sec-st.tv_sec)*1000*1000 + (et.tv_usec-st.tv_usec))

#ifdef DO_GCRLOG
int GC_gcrlog_put_mark GC_PROTO((int us));
int GC_gcrlog_put_sweep GC_PROTO((int us));
#endif

#ifdef DO_LDLOG
/*******************************************************************/

/* Should be called by one process. */
int GC_ldlog_init()
{
    GC_g_ldlogp = (struct g_ldlog *)
	GC_scratch_g_alloc(sizeof(struct g_ldlog));
    GC_g_ldlogp->phase = 0;
    GC_g_ldlogp->doing = 0;
    return 0;
}

/* called by each GC process. */
/* don't use GC_my_envp in this function. */
int GC_ldlog_init_local(GC_thread envp)
{
    GC_lg_ldlogp(envp) = (struct lg_ldlog *)
	GC_scratch_generic_alloc(sizeof(struct lg_ldlog),
				 &GC_l_scratch(envp));
    return 0;
}

/* Should be called by one process. */
int GC_ldlog_init_phase()
{
    struct timeval t;
    GC_thread envp;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_lg_ldlogp(envp)->index = 0;
    }
    GC_g_ldlogp->doing = 1;
    gettimeofday(&t, NULL);
    GC_g_ldlogp->st = t;
    return 0;
}

/* Should be called by each process. */
int GC_ldlog_put(SGC_MYENV_ARG_DECL1 int com, int kind)
{
    DECLARE_MYENV
    int i;
    struct timeval tv;
    if (GC_g_ldlogp->doing == 0) {
	return 0;
    }
    gettimeofday(&tv, NULL);
    i = GC_myg_ldlogp->index;
    if (i >= LDLOG_NUM) {
	PSM_mutex_lock(&GC_g_debug.print_ml);
	GC_err_printf3("pid %d FAILED TO PUT LDLOG(%d) [%s]...\n",
		GC_my_pid, i,
		GC_const_ldlog_comname[com]);
	PSM_mutex_unlock(&GC_g_debug.print_ml);
	return -1;
    }
    {
	struct ldlog_entry *p;
	p = &GC_myg_ldlogp->data[i];
	p->com = com;
	p->kind = kind;
	p->tv = tv;
    }
    i++;
    GC_myg_ldlogp->index = i;
    return 0;
}

/* Should be called by one process. */
int GC_ldlog_end_phase()
{
    GC_thread envp;
    int i;
    FILE *fp;
    char *prefn;
    char fn[64];
    int com;
    int kind;
    int pno;
    
    GC_g_ldlogp->doing = 0;
    prefn = getenv("GCLDLOG");
    if (prefn == NULL) {
	prefn = (char *)"ldlog";
    }
    sprintf(fn, "%s%d", prefn, GC_g_ldlogp->phase);
    fp = fopen(fn, "w");
    if (fp == NULL) {
	GC_err_printf0("FAILED TO WRITE LDLOG TO FILE...\n");
	return -1;
    }

    pno = 0;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	for (i = 0; i < GC_lg_ldlogp(envp)->index; i++) {
	    com = GC_lg_ldlogp(envp)->data[i].com;
	    kind = GC_lg_ldlogp(envp)->data[i].kind;
	    fprintf(fp, "%d %d %s %s\n",
		    SHZZZ_CALCDIFFTIME(GC_g_ldlogp->st,
				       GC_lg_ldlogp(envp)->data[i].tv),
		    pno,
		    GC_const_ldlog_comname[com],
		    GC_const_ldlog_kindname[kind]);
	}
	pno++;
    }

    fclose(fp);
    GC_g_ldlogp->phase++;
    GC_err_printf1("Wrote LDLOG to %s.\n", fn);
    return 0;
}
#endif

#ifdef DO_ALLOCLOG
/*******************************************************************/

/* Should be called by one process. */
int GC_alloclog_init()
{
    char *prefn;
    struct timeval t;
    char fn[64];

    GC_g_alloclogp = (struct g_alloclog *)
	GC_scratch_g_alloc(sizeof(struct g_alloclog));
    gettimeofday(&t, NULL);
    GC_g_alloclogp->st = t;

    prefn = getenv("GCALLOCLOG");
    if (prefn == NULL) {
	prefn = (char *)"alloclog";
    }
    sprintf(fn, "%s", prefn);
    BCOPY(fn, GC_g_alloclogp->fn, strlen(fn)+1);
    GC_g_alloclogp->fp = fopen(GC_g_alloclogp->fn, "w");
    if (GC_g_alloclogp->fp == NULL) {
	GC_err_printf0("FAILED TO WRITE ALLOCLOG TO FILE...\n");
	return -1;
    }

    PSM_mutex_init(&GC_g_alloclogp->ml);
    GC_g_alloclogp->pcount = 0;
    return 0;
}

/* called by each App process. */
/* don't use GC_my_envp in this function. */
int GC_alloclog_init_local(GC_thread envp)
{
    GC_la_alloclogp(envp) = (struct la_alloclog *)
	GC_scratch_generic_alloc(sizeof(struct la_alloclog),
				 &GC_l_scratch(envp));
    GC_la_alloclogp(envp)->index = 0;
    return 0;
}

/* Should be called by each process. */
int GC_alloclog_put(SGC_MYENV_ARG_DECL1 int com)
{
    DECLARE_MYENV
    int i;
    struct timeval tv;
    gettimeofday(&tv, NULL);
    i = GC_mya_alloclogp->index;
    if (i >= ALLOCLOG_NUM) {
	PSM_mutex_lock(&GC_g_debug.print_ml);
	GC_err_printf3("pid %d FAILED TO PUT ALLOCLOG(%d)...\n",
		GC_my_pid, i);
	PSM_mutex_unlock(&GC_g_debug.print_ml);
	return -1;
    }
    {
	struct alloclog_entry *p;
	p = &GC_mya_alloclogp->data[i];
	p->com = com;
	p->tv = tv;
    }
    i++;
    GC_mya_alloclogp->index = i;
    return 0;
}

/* Should be called by each App process. */
/* don't use GC_my_envp in this function. */
int GC_alloclog_end_local(GC_thread envp)
{
    int i;
    int com;
    
    PSM_mutex_lock(&GC_g_alloclogp->ml);

    for (i = 0; i < GC_la_alloclogp(envp)->index; i++) {
	com = GC_la_alloclogp(envp)->data[i].com;
	fprintf(GC_g_alloclogp->fp, "%d %d %s %s\n",
		SHZZZ_CALCDIFFTIME(GC_g_alloclogp->st,
				   GC_la_alloclogp(envp)->data[i].tv),
		GC_g_alloclogp->pcount,
		GC_const_alloclog_comname[com],
		"ALLOC");
    }

    GC_g_alloclogp->pcount++;
    PSM_mutex_unlock(&GC_g_alloclogp->ml);

    return 0;
}

/* Should be called by one App process. */
int GC_alloclog_end()
{
    fclose(GC_g_alloclogp->fp);
    GC_err_printf1("Wrote ALLOCLOG to %s.\n", GC_g_alloclogp->fn);
    return 0;
}
#endif

#ifdef DO_MARKLOG
/*******************************************************************/

/* Should be called by one GC process. */
int GC_marklog_init()
{
    GC_g_marklogp = (struct g_marklog *)
	GC_scratch_g_alloc(sizeof(struct g_marklog));
    return 0;
}

/* Should be called by each GC process. */
/* don't use GC_my_envp in this function. */
int GC_marklog_init_local(GC_thread envp)
{
    GC_l_marklogp(envp) = (struct l_marklog *)
	GC_scratch_generic_alloc(sizeof(struct l_marklog),
				 &GC_l_scratch(envp));
    return 0;
}

/* Should be called by one GC process. */
/* Currently, log data on GC processes are used, */
/* data on App processes are ignored. */
int GC_marklog_init_phase()
{
    int k;
    GC_thread envp;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	for (k = 0; k < MARKLOG_KIND; k++) {
	    GC_l_marklogp(envp)->count[k] = 0;
	}
    }
    GC_g_marklogp->currentclock = 0;
    return 0;
}

/* Should be called by one process. */
int GC_marklog_start_clock()
{
    struct timeval t;
    int c = GC_g_marklogp->currentclock;
    if (c >= MARKLOG_CLOCK) return -1;

    gettimeofday(&t, NULL);
    GC_g_marklogp->rsts[c] = t;
    return 0;
}

/* Should be called by each process. */
int GC_marklog_put(SGC_MYENV_ARG_DECL1 int kind, int count)
{
    DECLARE_MYENV
    GC_my_marklogp->count[kind] += count;
    return 0;
}

/* Should be called by one process. */
int GC_marklog_stop_clock()
{
    struct timeval t;
    int c = GC_g_marklogp->currentclock;
    gettimeofday(&t, NULL);
    GC_g_marklogp->rets[c] = t;
    GC_g_marklogp->currentclock++;
    return 0;
}

#ifdef SPARC
#if 1
/* for ISSUE (Enterprise10000) */
/* Ideal performance on 1pe would 1000. */
double const_a[] = { 
    0.50, 0.16, 0.020,  2.0, 1.3,
    0.4, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0,
};
#endif
#endif

#ifdef MIPS
/* for CHORUS (Origin2000) */
/* Ideal performance on 1pe would 1000. */
double const_a[] = {
    0.42, 0.13, 0.028,  2.0, 1.3,
    0.4, 0, 0, 0, 0,
    0, 0, 0 ,0, 0, 0, 0,
};
#endif

#ifdef I386
/* for GRACE (PC SMP) */
/* Ideal performance on 1pe would 1000. */
double const_a[] = {
    0.42, 0.13, 0.028,  2.0, 1.3,
    0.4, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0,
};
#endif

#ifdef HP_PA
/* for HP (???) */
/* Ideal performance on 1pe would 1000. */
double const_a[] = {
    0.42, 0.13, 0.028,  2.0, 1.3, /* ??? */
    0.4, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0,
};
#endif

/* Should be called by one process. */
int GC_marklog_end_phase()
{
    double sum_count[MARKLOG_KIND];
    int max_count[MARKLOG_KIND];
    int k;
    int npe;
    double ts[MARKLOG_CLOCK];
    double sum_t;
    int c;
    double sum_ans;
    double max_ans;
    char buf[256];
    GC_thread envp;

    sum_t = 0.0;
    for (c = 0; c < GC_g_marklogp->currentclock; c++) {
	ts[c] = (double)SHZZZ_CALCDIFFTIME(GC_g_marklogp->rsts[c],
					   GC_g_marklogp->rets[c])
	    /1000.0;
	sum_t += ts[c];
#if DO_MARKLOG > 1
	if (c > 0) {
	    sprintf(buf, "## %d -- %d interval %12.3f ms\n",
		    c-1, c, 
		    (double)SHZZZ_CALCDIFFTIME(GC_g_marklogp->rets[c-1],
					       GC_g_marklogp->rsts[c])
		    /1000.0);
	    GC_err_printf0(buf);
	}
	sprintf(buf, "## %d took %12.3f ms\n",
		c, ts[c]);
	GC_err_printf0(buf);
#endif
    }
#ifdef DO_GCRLOG
    GC_gcrlog_put_mark((int)(ts[0]*1000.0));
    GC_gcrlog_put_sweep((int)(ts[1]*1000.0));
#endif

    for (k = 0; k < MARKLOG_KIND; k++) {
	sum_count[k] = 0.0;
	max_count[k] = 0;
    }

    npe = 0;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	for (k = 0; k < MARKLOG_KIND; k++) {
	    long cl = GC_l_marklogp(envp)->count[k];
	    sum_count[k] += cl;
	    if (cl > max_count[k]) {
		max_count[k] = cl;
	    }
	}
	npe++;
    }

    /* All Stat */
    sprintf(buf, "#STAT--# ----- MARKLOG START -----\n");
    GC_err_printf0(buf);
    sum_ans = 0.0;
    max_ans = 0.0;
    for (k = 0; k < MARKLOG_KIND; k++) {
#if DO_MARKLOG > 1
	sprintf(buf, "#STAT%02d#   %12.0f %9d %12.3f   %12.3f %d\n",
		k,
		sum_count[k],
		max_count[k],
		sum_t,
		(double)sum_count[k]/sum_t,
		npe);
	GC_err_printf0(buf);
#endif
	sum_ans += const_a[k] * (double)sum_count[k];
	max_ans += const_a[k] * (double)max_count[k];
    }

    /* This includes load balancing workload. */
#if DO_MARKLOG > 1
    sprintf(buf, "#STATZA#   %12.0f %9d %12.3f   %12.3f %d\n",
	    sum_ans,
	    (int)max_ans,
	    sum_t,
	    (double)sum_ans/(double)sum_t,
	    npe);
    GC_err_printf0(buf);
#endif

    /* Mark Only */
    sum_ans = 0.0;
    max_ans = 0.0;
    for (k = 0; k <= 2; k++) {
	sum_ans += const_a[k] * (double)sum_count[k];
	max_ans += const_a[k] * (double)max_count[k];
    }
    sprintf(buf, "#STATZM#   %12.0f %9d %12.3f   %12.3f %d\n",
	    sum_ans,
	    (int)max_ans,
	    ts[0],
	    (double)sum_ans/ts[0],
	    npe);
    GC_err_printf0(buf);
    
    /* Sweep Only */
    sum_ans = 0.0;
    max_ans = 0.0;
    for (k = 3; k <= 4; k++) {
	sum_ans += const_a[k] * (double)sum_count[k];
	max_ans += const_a[k] * (double)max_count[k];
    }
    sprintf(buf, "#STATZS#   %12.0f %9d %12.3f   %12.3f %d\n",
	    sum_ans,
	    (int)max_ans,
	    ts[1],
	    (double)sum_ans/ts[1],
	    npe);
    GC_err_printf0(buf);
    
    /* Mark+Sweep */
    sum_ans = 0.0;
    max_ans = 0.0;
    for (k = 0; k <= 5; k++) {
	sum_ans += const_a[k] * (double)sum_count[k];
	max_ans += const_a[k] * (double)max_count[k];
    }
    sprintf(buf, "#STATZZ#   %12.0f %9d %12.3f   %12.3f %d\n",
	    sum_ans,
	    (int)max_ans,
	    sum_t,
	    (double)sum_ans/sum_t,
	    npe);
    GC_err_printf0(buf);
    sprintf(buf, "#STAT--# ----- MARKLOG END -----\n");
    GC_err_printf0(buf);

    GC_g_marklogp->currentclock = -1;
    return 0;
}

#endif

#ifdef DO_PICLOG
/*******************************************************************/

#ifdef SPARC
/* Description of event No is in picio.h. */
#  define GC_PIC_EVENT0 PICEVENT0_EC_REF
#  define GC_PIC_EVENT1 PICEVENT1_EC_HIT
#endif

#ifdef MIPS
/* Description of event No is in /usr/include/sys/hwperfmacros.h. */
#  define GC_PIC_EVENT0 HWPERF_C0PRFCNT0_CYCLES
#  define GC_PIC_EVENT1 HWPERF_C0PRFCNT1_SDCMISS
#endif

/* Should be called by one process. */
int GC_piclog_init()
{
    GC_g_piclogp = (struct g_piclog *)
	GC_scratch_g_alloc(sizeof(struct g_piclog));
    return 0;
}

/* Should be called by each GC process. */
/* don't use GC_my_envp in this function. */
int GC_piclog_init_local(GC_thread envp)
{
    GC_l_piclogp(envp) = (struct l_piclog *)
	GC_scratch_generic_alloc(sizeof(struct l_piclog),
				 &GC_l_scratch(envp));
    return 0;
}

int GC_piclog_init_phase_local(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int mode;
    int c;
    PERFC_counter_t pic[2];

    for (mode = 0; mode < PICLOG_MODE; mode++) {
	for (c = 0; c < 2; c++) {
	    GC_my_piclogp->count[mode][c] = 0;
	}
    }
    GC_my_piclogp->mode = 0;

#if defined PROCESSOR_BIND_SUPPORTED && !defined DO_PROCESSOR_BIND
    /*GC_sh_processor_bind_currentpe();*/
#endif
    if (PERFC_open(&GC_my_piclogp->handler,
		   GC_PIC_EVENT0, GC_PIC_EVENT1) < 0) {
    }

    if (PERFC_read(&GC_my_piclogp->handler, &pic[0], &pic[1]) < 0) {
	GC_my_piclogp->start[0] = (long)0;
	GC_my_piclogp->start[1] = (long)0;
	return -1;
    }
    GC_my_piclogp->start[0] = (long)pic[0];
    GC_my_piclogp->start[1] = (long)pic[1];

    return 0;
}

int GC_piclog_change_mode(SGC_MYENV_ARG_DECL1 int mode)
{
    DECLARE_MYENV
    int oldmode;
    PERFC_counter_t pic0, pic1;

    if (PERFC_read(&GC_my_piclogp->handler,
		   &pic0, &pic1) < 0) {
	return -1;
    }

    if (mode >= PICLOG_MODE) {
	GC_err_printf1("piclog_change_mode: ERROR! mode %d\n", mode);
    }
    oldmode = GC_my_piclogp->mode;
    if (mode != oldmode) {
	if (oldmode >= 0) {
	    GC_my_piclogp->count[oldmode][0] +=
		(long)pic0 - GC_my_piclogp->start[0];
	    GC_my_piclogp->count[oldmode][1] +=
		(long)pic1 - GC_my_piclogp->start[1];
	}

	GC_my_piclogp->mode = mode;
	if (mode >= 0) {	
	    if (PERFC_read(&GC_my_piclogp->handler,
			   &pic0, &pic1) < 0) {
		return -1;
	    }
	    GC_my_piclogp->start[0] = (long)pic0;
	    GC_my_piclogp->start[1] = (long)pic1;
	}
    }
    return 0;
}

int GC_piclog_add(SGC_MYENV_ARG_DECL1 int mode, int c0, int c1)
{
    DECLARE_MYENV
    if (c0 < 0 || c1 < 0) return -1;
    if (c0 > 0x10000 || c1 > 0x10000) return -1;
    if (mode >= PICLOG_MODE) {
	GC_err_printf1("piclog_add: ERROR! mode %d\n", mode);
    }
    GC_my_piclogp->count[mode][0] += c0;
    GC_my_piclogp->count[mode][1] += c1;
    return 0;
}

int GC_piclog_end_phase_local(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    /* Stop measuring */
    GC_piclog_change_mode(SGC_MYENV_ARG1 -1);
    PERFC_close(&GC_my_piclogp->handler);
#if defined PROCESSOR_BIND_SUPPORTED && !defined DO_PROCESSOR_BIND
    /*GC_sh_processor_unbind();*/
#endif
    return 0;
}

/* Should be called by one process. */
int GC_piclog_end_phase()
{
    int mode;
    int c;
    int sum_count[PICLOG_MODE][2];
    int max_count[PICLOG_MODE][2];
    GC_thread envp;

    for (mode = 0; mode < PICLOG_MODE; mode++) {
	for (c = 0; c < 2; c++) {
	    sum_count[mode][c] = 0;
	    max_count[mode][c] = 0;
	}
    }

    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	for (mode = 0; mode < PICLOG_MODE; mode++) {
	    for (c = 0; c < 2; c++) {
		int data = GC_l_piclogp(envp)->count[mode][c];
		sum_count[mode][c] += data;
		if (data > max_count[mode][c]) {
		    max_count[mode][c] = data;
		}
	    }
	}
    }

    GC_err_printf0("#PIC--# PICLOG: sum0 sum1 max0 max1\n");
    for (mode = 0; mode < PICLOG_MODE; mode++) {
	GC_err_printf5("#PIC%02d#  %12ld %12ld %12ld %12ld\n",
		       mode, sum_count[mode][0], sum_count[mode][1],
		       max_count[mode][0], max_count[mode][1]);
    }

#if 0
    for (mode = 0; mode < PICLOG_MODE; mode += 2) {
	GC_err_printf4("#PIC%02d-%02d# %12ld %12ld\n",
		mode, mode+1,
		sum_count[mode][0]-sum_count[mode+1][0],
		sum_count[mode][1]-sum_count[mode+1][1]);
    }
#endif

    return 0;
}

#endif

#ifdef DO_GCRLOG
/*******************************************************************/
/* Should be called by one process, but several call is OK. */
int GC_gcrlog_init()
{
    struct timeval t;
    GC_g_gcrlogp = (struct g_gcrlog *)
	GC_scratch_g_alloc(sizeof(struct g_gcrlog));
    gettimeofday(&t, NULL);
    GC_g_gcrlogp->st = t;
    GC_g_gcrlogp->gc_us = 0;
    GC_g_gcrlogp->sweep_us = 0;
    GC_g_gcrlogp->mark_us = 0;
    return 0;
}

int GC_gcrlog_end()
{
    int markms;
    int gcms;
    int sweepms;
    int appms;
    struct timeval st;
    struct timeval et;

    gettimeofday(&et, NULL);
    st = GC_g_gcrlogp->st;

    appms = ((et.tv_sec-st.tv_sec)*1000 + 
	     (et.tv_usec-st.tv_usec+500)/1000);
    gcms = (GC_g_gcrlogp->gc_us + 500)/1000;
    markms = (GC_g_gcrlogp->mark_us + 500)/1000;
    sweepms = (GC_g_gcrlogp->sweep_us + 500)/1000;

    PSM_mutex_lock(&GC_g_debug.print_ml);
    GC_err_printf2("GCRLOG: GC_gc_no = %d, GC_heapsize = %dk\n",
		   GC_gc_no, GC_heapsize/1024);
    GC_err_printf4("GCRLOG: MARK %d ms / SWEEP %d ms / GC %d ms / ALL %d ms \n",
	    markms, sweepms, gcms, appms);
    PSM_mutex_unlock(&GC_g_debug.print_ml);

    return 0;
}

int GC_gcrlog_put_mark(int us)
{
    GC_g_gcrlogp->mark_us += us;
    return 0;
}

int GC_gcrlog_put_sweep(int us)
{
    GC_g_gcrlogp->sweep_us += us;
    return 0;
}

int GC_gcrlog_init_gc()
{
    struct timeval t;
    gettimeofday(&t, NULL);
    GC_g_gcrlogp->gc_st = t;
    return 0;
}

void GC_gcrlog_set_clock(int i)
{
    struct timeval t;
    gettimeofday(&t, NULL);
    GC_g_gcrlogp->clocks[i] = t;
}

int GC_gcrlog_end_gc()
{
    struct timeval st;
    struct timeval et;
    int i;

    gettimeofday(&et, NULL);
    st = GC_g_gcrlogp->gc_st;
    GC_g_gcrlogp->gc_us += SHZZZ_CALCDIFFTIME(st, et);

    GC_err_printf0("GCRLOG: (init: marksweep: finalize) = (");
    for (i = 1; i < 4; i++) {
	int span;
	span = SHZZZ_CALCDIFFTIME(GC_g_gcrlogp->clocks[i-1],
				  GC_g_gcrlogp->clocks[i]) / 1000;
	GC_err_printf1("%d ms: ", span);
    }
    GC_err_printf0(")\n");

    return 0;
}

#endif

#ifdef DO_MSLOG
/*******************************************************************/

/* Should be called by one process. */
int GC_mslog_init()
{
    GC_g_mslogp = (struct g_mslog *)
	GC_scratch_g_alloc(sizeof(struct g_mslog));
    GC_g_mslogp->phase = 0;
    GC_g_mslogp->doing = 0;
    return 0;
}

/* called by each GC process. */
/* don't use GC_my_envp in this function. */
int GC_mslog_init_local(GC_thread envp)
{
    GC_lg_mslogp(envp) = (struct lg_mslog *)
	GC_scratch_generic_alloc(sizeof(struct lg_mslog),
				 &GC_l_scratch(envp));
    return 0;
}

/* Should be called by one process. */
int GC_mslog_init_phase()
{
    GC_thread envp;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_lg_mslogp(envp)->index = 0;
	GC_lg_mslogp(envp)->time = 0;
    }
    GC_g_mslogp->doing = 1;
    return 0;
}

/* Should be called by each process. */
int GC_mslog_put(SGC_MYENV_ARG_DECL1 int com, word *obj)
{
    DECLARE_MYENV
    int i;
    int t;
    struct mslog_entry *p;
    if (GC_g_mslogp->doing == 0) {
	return 0;
    }
    i = GC_myg_mslogp->index;
    t = GC_myg_mslogp->time;
    if (i >= MSLOG_NUM) {
	PSM_mutex_lock(&GC_g_debug.print_ml);
	GC_err_printf3("pid %d FAILED TO PUT MSLOG(%d) [%s]...\n",
		GC_my_pid, i,
		GC_const_mslog_comname[com]);
	PSM_mutex_unlock(&GC_g_debug.print_ml);
	return -1;
    }
    if (i > 0 && com == GC_MSC_POP &&
	GC_myg_mslogp->data[i-1].obj == obj) {
	p = &GC_myg_mslogp->data[i-1]; /* previous data */
	if (p->com == GC_MSC_PUSH) {
	    /* change previous command */
	    p->com = GC_MSC_PUSH_POP;
	}
	else if (p->com == GC_MSC_STEAL) {
	    /* change previous command */
	    p->com = GC_MSC_STEAL_POP;
	}
	else {
	    PSM_mutex_lock(&GC_g_debug.print_ml);
	    GC_err_printf3("pid %d UNKNOWN TYPE TO PUT MSLOG(%d) [%s]...\n",
			   GC_my_pid, i,
			   GC_const_mslog_comname[com]);
	    PSM_mutex_unlock(&GC_g_debug.print_ml);
	    return -1;
	}
    }
    else {
	p = &GC_myg_mslogp->data[i];
	p->timestamp = t;
	p->com = com;
	p->obj = obj;
	GC_myg_mslogp->index = i+1;
    }
    GC_myg_mslogp->time = t+1;
    return 0;
}

/* Should be called by one process. */
int GC_mslog_end_phase()
{
    GC_thread envp;
    int i;
    FILE *fp;
    char *prefn;
    char fn[64];
    int com;
    word *obj;
    int timestamp;
    int procno;
    
    GC_g_mslogp->doing = 0;
    prefn = getenv("GCMSLOG");
    if (prefn == NULL) {
	prefn = (char *)"mslog";
    }
    sprintf(fn, "%s%d", prefn, GC_g_mslogp->phase);
    fp = fopen(fn, "w");
    if (fp == NULL) {
	GC_err_printf0("FAILED TO WRITE MSLOG TO FILE...\n");
	return -1;
    }

    procno = 0;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	for (i = 0; i < GC_lg_mslogp(envp)->index; i++) {
	    com = GC_lg_mslogp(envp)->data[i].com;
	    obj = GC_lg_mslogp(envp)->data[i].obj;
	    timestamp = GC_lg_mslogp(envp)->data[i].timestamp;
	    fprintf(fp, "%d %d %s %08x\n",
		    timestamp,
		    procno,
		    GC_const_mslog_comname[com],
		    obj);
	}
	procno++;
    }

    fclose(fp);
    GC_g_mslogp->phase++;
    GC_err_printf1("Wrote MSLOG to %s.\n", fn);
    return 0;
}
#endif


#ifdef DO_CACHELOG
/*******************************************************************/

/* called by each GC process. */
/* don't use GC_my_envp in this function. */
int GC_cachelog_init_local(GC_thread envp)
{
    GC_l_cachelogp(envp) = (struct l_cachelog *)
	GC_scratch_generic_alloc(sizeof(struct l_cachelog),
				 &GC_l_scratch(envp));
    return 0;
}

/* Should be called by one process. */
int GC_cachelog_init_phase()
{
    GC_thread envp;
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_l_cachelogp(envp)->prevhdr = NULL;
	GC_l_cachelogp(envp)->mbhit = 0;
	GC_l_cachelogp(envp)->mbmiss = 0;
	GC_l_cachelogp(envp)->objmiss = 0;
    }
    return 0;
}

/* Should be called by each process. */
int GC_cachelog_before_loop(SGC_MYENV_ARG_DECL0)
{
    DECLARE_MYENV

    GC_my_cachelogp->prevhdr = NULL;
    return 0;
}

/* Should be called by each process. */
int GC_cachelog_put_mbhdr(SGC_MYENV_ARG_DECL1 hdr *hdr)
{
    DECLARE_MYENV

    if (hdr == NULL) {
	GC_my_cachelogp->prevhdr = NULL;
	return 0;
    }

    if (hdr == GC_my_cachelogp->prevhdr) {
	GC_my_cachelogp->mbhit++;
    }
    else {
	GC_my_cachelogp->mbmiss++;
	GC_my_cachelogp->prevhdr = hdr;
    }
    return 0;
}

/* Should be called by each process. */
int GC_cachelog_put_objsize(SGC_MYENV_ARG_DECL1 int sizew)
{
    DECLARE_MYENV

    int linesize = MARK_BITS_SZ;
    GC_my_cachelogp->objmiss += (sizew + linesize - 1) / linesize;
    return 0;
}

int GC_cachelog_end_phase()
{
    GC_thread envp;
    int summbhit = 0;
    int summbmiss = 0;
    int sumobjmiss = 0;

    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;

	summbhit += GC_l_cachelogp(envp)->mbhit;
	summbmiss += GC_l_cachelogp(envp)->mbmiss;
	sumobjmiss += GC_l_cachelogp(envp)->objmiss;
    }
    GC_err_printf0("CACHELOG: mb-hit   mb-miss  obj-miss\n");
    GC_err_printf3("CACHELOG: %8d %8d %8d\n",
		   summbhit, summbmiss, sumobjmiss);
    return 0;
}

#endif
